# -*- coding: utf-8 -*-

import itertools
import logging
import numpy as np
from numpy.lib import (
    recfunctions as np_rfn,
)
from os import (
    makedirs,
)
import os.path
from scipy.interpolate import (
    interp1d,
)
from scipy.stats import (
    gamma,
)

try:
    from iminuit import minimize
except Exception:
    IMINUIT_LOADED = False
else:
    IMINUIT_LOADED = True

from skyllh.core.progressbar import (
    ProgressBar,
)
from skyllh.core.py import (
    float_cast,
    int_cast,
    issequence,
    issequenceof,
)
from skyllh.core.session import (
    is_interactive_session,
)
from skyllh.core.source_model import (
    PointLikeSource,
)
from skyllh.core.storage import (
    NPYFileLoader,
)
from skyllh.core.utils.spline import (
    make_spline_1d,
)


"""This module contains common utility functions useful for an analysis.
"""


def pointlikesource_to_data_field_array(
        tdm, shg_mgr, pmm):
    """Function to transform a list of PointLikeSource sources into a numpy
    record ndarray. The resulting numpy record ndarray contains the following
    fields:

        `ra`: float
            The right-ascention of the point-like source.
        `dec`: float
            The declination of the point-like source.
        `weight`: float
            The weight of the point-like source.

    Parameters
    ----------
    tdm : instance of TrialDataManager
        The TrialDataManager instance.
    shg_mgr : instance of SourceHypoGroupManager
        The instance of SourceHypoGroupManager that defines the sources.
    pmm : instance of ParameterModelMapper
        The instance of ParameterModelMapper that defines the mapping of global
        parameters to local model parameters.

    Returns
    -------
    arr : (N_sources,)-shaped numpy record ndarray
        The numpy record ndarray holding the source parameters.
    """
    sources = shg_mgr.source_list

    if not issequenceof(sources, PointLikeSource):
        raise TypeError(
            'The sources of the SourceHypoGroupManager must be '
            'PointLikeSource instances!')

    arr = np.empty(
        (len(sources),),
        dtype=[
            ('ra', np.float64),
            ('dec', np.float64),
            ('weight', np.float64),
        ],
        order='F')

    for (i, src) in enumerate(sources):
        arr['ra'][i] = src.ra
        arr['dec'][i] = src.dec
        arr['weight'][i] = src.weight

    return arr


def calculate_pval_from_trials(
        ts_vals, ts_threshold, comp_operator='greater'):
    """Calculates the percentage (p-value) of test-statistic trials that are
    above the given test-statistic critical value.
    In addition it calculates the standard deviation of the p-value assuming
    binomial statistics.

    Parameters
    ----------
    ts_vals : (n_trials,)-shaped 1D ndarray of float
        The ndarray holding the test-statistic values of the trials.
    ts_threshold : float
        The critical test-statistic value.
    comp_operator: string, optional
        The comparison operator for p-value calculation. It can be set to one of
        the following options: 'greater' or 'greater_equal'.

    Returns
    -------
    p, p_sigma: tuple(float, float)
    """
    if comp_operator == 'greater':
        p = ts_vals[ts_vals > ts_threshold].size / ts_vals.size
    elif comp_operator == 'greater_equal':
        p = ts_vals[ts_vals >= ts_threshold].size / ts_vals.size
    else:
        raise ValueError(
            f"The comp_operator={comp_operator} is not an"
            "available option ('greater' or 'greater_equal')."
        )

    p_sigma = np.sqrt(p * (1 - p) / ts_vals.size)

    return (p, p_sigma)


def calculate_pval_from_gammafit_to_trials(
        ts_vals,
        ts_threshold,
        eta=3.0,
        n_max=500000):
    """Calculates the probability (p-value) of test-statistic exceeding
    the given test-statistic threshold. This calculation relies on fitting
    a gamma distribution to a list of ts values.

    Parameters
    ----------
    ts_vals : (n_trials,)-shaped 1D ndarray of float
        The ndarray holding the test-statistic values of the trials.
    ts_threshold : float
        The critical test-statistic value.
    eta : float, optional
        Test-statistic value at which the gamma function is truncated
        from below. Default = 3.0.
    n_max : int, optional
        The maximum number of trials that should be used during
        fitting. Default = 500,000

    Returns
    -------
    p, p_sigma: tuple(float, float)
    """
    if not IMINUIT_LOADED:
        raise ImportError(
            'The iminuit module was not imported! '
            'This module is a requirement for the function '
            '"calculate_pval_from_gammafit_to_trials"!')

    if ts_threshold < eta:
        raise ValueError(
            'ts threshold value = %e, eta = %e. The calculation of the p-value'
            'from the fit is correct only for ts threshold larger than '
            'the truncation threshold eta.',
            ts_threshold, eta)

    if len(ts_vals) > n_max:
        ts_vals = ts_vals[:n_max]

    Ntot = len(ts_vals)
    ts_eta = ts_vals[ts_vals > eta]
    N_prime = len(ts_eta)
    alpha = N_prime/Ntot

    def obj(x):
        return truncated_gamma_logpdf(
            x[0],
            x[1],
            eta=eta,
            ts_above_eta=ts_eta,
            N_above_eta=N_prime)

    x0 = [0.75, 1.8]  # Initial values of function parameters.
    bounds = [[0.1, 10], [0.1, 10]]  # Ranges for the minimization fitter.
    r = minimize(obj, x0, bounds=bounds)
    pars = r.x

    norm = alpha/gamma.sf(eta, a=pars[0], scale=pars[1])
    p = norm * gamma.sf(ts_threshold, a=pars[0], scale=pars[1])

    # a correct calculation of the error in pvalue due to
    # fitting uncertainty remains to be implemented
    # return p_sigma = 0 for now for consistentcy with
    # calculate_pval_from_trials()
    p_sigma = 0.0
    return (p, p_sigma)


def calculate_pval_from_trials_mixed(
        ts_vals,
        ts_threshold,
        switch_at_ts=3.0,
        eta=None,
        n_max=500000,
        comp_operator='greater_equal'):
    """Calculates the probability (p-value) of test-statistic exceeding
    the given test-statistic threshold. This calculation relies on fitting
    a gamma distribution to a list of ts values if ts_threshold is larger than
    switch_at_ts. If ts_threshold is smaller then the pvalue will be taken
    from the trials directly.

    Parameters
    ----------
    ts_vals : (n_trials,)-shaped 1D ndarray of float
        The ndarray holding the test-statistic values of the trials.
    ts_threshold : float
        The critical test-statistic value.
    switch_at_ts : float, optional
        Test-statistic value below which p-value is computed from trials
        directly. For thresholds greater than switch_at_ts the pvalue is
        calculated using a gamma fit.
    eta : float, optional
        Test-statistic value at which the gamma function is truncated
        from below. Default is None.
    n_max : int, optional
        The maximum number of trials that should be used during
        fitting. Default = 500,000
    comp_operator: string, optional
        The comparison operator for p-value calculation. It can be set to one of
        the following options: 'greater' or 'greater_equal'.

    Returns
    -------
    p, p_sigma: tuple(float, float)
    """
    # Set `eta` to `switch_at_ts` as a default.
    # It makes sure that both functions return the same pval at `switch_at_ts`.
    if eta is None:
        eta = switch_at_ts

    if ts_threshold < switch_at_ts:
        return calculate_pval_from_trials(
            ts_vals,
            ts_threshold,
            comp_operator=comp_operator)
    else:
        return calculate_pval_from_gammafit_to_trials(
            ts_vals,
            ts_threshold,
            eta=eta,
            n_max=n_max)


def truncated_gamma_logpdf(
        a,
        scale,
        eta,
        ts_above_eta,
        N_above_eta):
    """Calculates the -log(likelihood) of a sample of random numbers
    generated from a gamma pdf truncated from below at x=eta.

    Parameters
    ----------
    a : float
        Shape parameter.
    scale : float
        Scale parameter.
    eta : float
        Test-statistic value at which the gamma function is truncated
        from below.
    ts_above_eta : (n_trials,)-shaped 1D ndarray
        The ndarray holding the test-statistic values falling in
        the truncated gamma pdf.
    N_above_eta : int
        Number of test-statistic values falling in the truncated
        gamma pdf.

    Returns
    -------
    -logl : float
    """
    c0 = 1. - gamma.cdf(eta, a=a, scale=scale)
    c0 = 1./c0
    logl = N_above_eta*np.log(c0)
    logl += np.sum(
        gamma.logpdf(
            ts_above_eta,
            a=a,
            scale=scale))

    return -logl


def calculate_critical_ts_from_gamma(
        ts,
        h0_ts_quantile,
        eta=3.0):
    """Calculates the critical test-statistic value corresponding
    to h0_ts_quantile by fitting the ts distribution with a truncated
    gamma function.

    Parameters
    ----------
    ts : (n_trials,)-shaped 1D ndarray
        The ndarray holding the test-statistic values of the trials.
    h0_ts_quantile : float
        Null-hypothesis test statistic quantile.
    eta : float, optional
        Test-statistic value at which the gamma function is truncated
        from below.

    Returns
    -------
    critical_ts : float
    """
    if not IMINUIT_LOADED:
        raise ImportError(
            'The iminuit module was not imported! '
            'This module is a requirement of the function '
            '"calculate_critical_ts_from_gamma"!')

    Ntot = len(ts)
    ts_eta = ts[ts > eta]
    N_prime = len(ts_eta)
    alpha = N_prime/Ntot

    def obj(x):
        return truncated_gamma_logpdf(
            x[0],
            x[1],
            eta=eta,
            ts_above_eta=ts_eta,
            N_above_eta=N_prime)

    x0 = [0.75, 1.8]  # Initial values of function parameters.
    bounds = [[0.1, 10], [0.1, 10]]  # Ranges for the minimization fitter.
    r = minimize(obj, x0, bounds=bounds)
    pars = r.x

    norm = alpha/gamma.sf(eta, a=pars[0], scale=pars[1])
    critical_ts = gamma.ppf(1 - 1./norm*h0_ts_quantile, a=pars[0], scale=pars[1])

    if critical_ts < eta:
        raise ValueError(
            'Critical ts value = %e, eta = %e. The calculation of the critical '
            'ts value from the fit is correct only for critical ts larger than '
            'the truncation threshold eta.',
            critical_ts, eta)

    return critical_ts


def polynomial_fit(
        ns,
        p,
        p_weight,
        deg,
        p_thr):
    """Performs a polynomial fit on the p-values of test-statistic trials
    associated to each ns..
    Using the fitted parameters it computes the number of signal events
    correponding to the given p-value critical value.

    Parameters
    ----------
    ns : 1D array_like object
        x-coordinates of the sample.
    p : 1D array_like object
        y-coordinates of the sample.
    p_weight : 1D array_like object
        Weights to apply to the y-coordinates of the sample points. For gaussian
        uncertainties, use 1/sigma.
    deg : int
        Degree of the fitting polynomial function.
    p_thr : float within [0,1]
        The critical p-value.

    Returns
    -------
    ns : float
    """
    (params, cov) = np.polyfit(ns, p, deg, w=p_weight, cov=True)

    # Check if the second order coefficient is positive and eventually
    # change to a polynomial fit of order 1 to avoid to overestimate
    # the mean number of signal events for the chosen ts quantile.
    if deg == 2 and params[0] > 0:
        deg = 1
        (params, cov) = np.polyfit(ns, p, deg, w=p_weight, cov=True)

    if deg == 1:
        (a, b) = (params[0], params[1])
        ns = (p_thr - b)/a
        return ns

    elif deg == 2:
        (a, b, c) = (params[0], params[1], params[2])
        ns = (- b + np.sqrt((b**2)-4*a*(c-p_thr))) / (2*a)
        return ns

    else:
        raise ValueError(
            'deg = %g is not valid. The order of the polynomial function '
            'must be 1 or 2.',
            deg)


def estimate_mean_nsignal_for_ts_quantile(  # noqa: C901
        ana,
        rss,
        p,
        eps_p,
        mu_range,
        critical_ts=None,
        h0_trials=None,
        h0_ts_quantile=None,
        min_dmu=0.5,
        bkg_kwargs=None,
        sig_kwargs=None,
        ppbar=None,
        tl=None,
        pathfilename=None):
    """Calculates the mean number of signal events needed to be injected to
    reach a test statistic distribution with defined properties for the given
    analysis.

    Parameters
    ----------
    ana : Analysis instance
        The Analysis instance to use for the calculation.
    rss : instance of RandomStateService
        The RandomStateService instance to use for generating random numbers.
    p : float
        Desired probability of signal test statistic for exceeding
        `h0_ts_quantile` part of null-hypothesis test statistic threshold.
    eps_p : float
        Precision in `p` as stopping condition for the calculation.
    mu_range : 2-element sequence
        The range of mu (lower,upper) to search for mean number of signal
        events.
    critical_ts : float | None
        The critical test-statistic value that should be overcome by the signal
        distribution. If set to None, the null-hypothesis test-statistic
        distribution will be used to compute the critical TS value.
    h0_trials : (n_h0_trials,)-shaped ndarray | None
        The structured ndarray holding the trials for the null-hypothesis.
        If set to `None`, the number of trials is calculated
        from binomial statistics via `h0_ts_quantile*(1-h0_ts_quantile)/eps**2`,
        where `eps` is `min(5e-3, h0_ts_quantile/10)`.
    h0_ts_quantile : float | None
        Null-hypothesis test statistic quantile.
        If set to None, the critical test-statistic value that should be
        overcome by the signal distribution MUST be given.
    min_dmu : float
        The minimum delta mu to use for calculating the derivative dmu/dp.
        The default is ``0.5``.
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    sig_kwargs : dict | None
        Additional keyword arguments for the `generate_signal_events` method
        of the `SignalGenerator` class. An usual keyword argument is
        `poisson`. If `poisson` is set to True, the actual number of
        generated signal events will be drawn from a Poisson distribution
        with the mean number of signal events, mu.
    ppbar : instance of ProgressBar | None
        The possible parent ProgressBar instance.
    tl: instance of TimeLord | None
        The optional TimeLord instance that should be used to collect timing
        information about this function.
    pathfilename: string | None
        Trial data file path including the filename.
        If set to None, generatedtrials won't be saved.

    Returns
    -------
    mu : float
        Estimated mean number of signal events.
    mu_err : None
        Error estimate needs to be implemented.
    """
    logger = logging.getLogger(__name__)

    n_total_generated_trials = 0

    if (critical_ts is None) and (h0_ts_quantile is None):
        raise RuntimeError(
            "Both the critical test-statistic value and the null-hypothesis "
            "test-statistic quantile are set to None. One of the two is "
            "needed to have the critical test-statistic value that defines "
            "the type of test to run.")
    elif critical_ts is None:
        n_trials_max = int(5.e5)
        # Via binomial statistics, calcuate the minimum number of trials
        # needed to get the required precision on the critial TS value.
        eps = min(0.005, h0_ts_quantile/10)
        n_trials_min = int(h0_ts_quantile*(1-h0_ts_quantile)/eps**2 + 0.5)

        # Compute either n_trials_max or n_trials_min trials depending on
        # which one is smaller. If n_trials_max trials are computed, a
        # fit to the ts distribution is performed to get the critial TS.
        n_trials_total = min(n_trials_min, n_trials_max)
        if h0_trials is None:
            h0_ts_vals = ana.do_trials(
                rss=rss,
                n=n_trials_total,
                mean_n_sig=0,
                bkg_kwargs=bkg_kwargs,
                sig_kwargs=sig_kwargs,
                ppbar=ppbar,
                tl=tl)['ts']

            logger.debug(
                'Generate %d null-hypothesis trials',
                n_trials_total)
            n_total_generated_trials += n_trials_total

            if pathfilename is not None:
                dirname = os.path.dirname(pathfilename)
                if dirname:
                    # Create the directory if dirname is not empty.
                    makedirs(dirname, exist_ok=True)
                # Save the trial data to file.
                np.save(pathfilename, h0_ts_vals)
        else:
            if h0_trials.size < n_trials_total:
                if not ('seed' in h0_trials.dtype.names):
                    logger.debug(
                        'Uploaded trials miss the rss_seed field. '
                        'Will not be possible to extend the trial file '
                        'safely. Uploaded trials will *not* be used.')
                    n_trials = n_trials_total
                    h0_ts_vals = ana.do_trials(
                        rss=rss,
                        n=n_trials,
                        mean_n_sig=0,
                        bkg_kwargs=bkg_kwargs,
                        sig_kwargs=sig_kwargs,
                        ppbar=ppbar,
                        tl=tl)['ts']
                else:
                    n_trials = n_trials_total - h0_trials.size
                    h0_ts_vals = extend_trial_data_file(
                        ana,
                        rss,
                        n_trials,
                        trial_data=h0_trials,
                        mean_n_sig=0,
                        pathfilename=pathfilename)['ts']
                logger.debug(
                    'Generate %d null-hypothesis trials',
                    n_trials)
                n_total_generated_trials += n_trials
            else:
                h0_ts_vals = h0_trials['ts']

        h0_ts_vals = h0_ts_vals[np.isfinite(h0_ts_vals)]
        logger.debug(
            'Number of trials after finite cut: %d',
            len(h0_ts_vals))
        logger.debug(
            'Min / Max h0 TS value: %e / %e',
            np.min(h0_ts_vals), np.max(h0_ts_vals))

        # If the minimum number of trials needed to get the required precision
        # on the critical TS value is smaller then 500k, compute the critical ts
        # value directly from trials; otherwise calculate it from the gamma
        # function fitted to the ts distribution.
        if n_trials_min <= n_trials_max:
            c = np.percentile(h0_ts_vals, (1 - h0_ts_quantile)*100)
        else:
            c = calculate_critical_ts_from_gamma(h0_ts_vals, h0_ts_quantile)
        logger.debug(
            'Critical ts value for bkg ts quantile %g: %e',
            h0_ts_quantile, c)
    elif h0_ts_quantile is None:
        # Make sure that the critical ts is a float.
        if not isinstance(critical_ts, float):
            raise TypeError(
                "The critical test-statistic value must be a float, not "
                f"{type(critical_ts)}!"
            )
        c = critical_ts
        logger.debug(
            'Critical ts value for upper limit: %e',
            c)
    else:
        raise RuntimeError(
            "Both a critical ts value and a null-hypothesis test_statistic "
            "quantile were given. If you want to use your critical_ts "
            "value, set h0_ts_quantile to None; if you want to compute the "
            "critical ts from the background distribution, set critical_ts "
            "to None.")

    # Make sure ns_range is mutable.
    ns_range_ = list(mu_range)

    ns_lower_bound = 0
    ns_upper_bound = +np.inf

    # The number of required trials per mu point for the desired uncertainty in
    # probability can be estimated via binomial statistics.
    n_trials = int(p*(1-p)/eps_p**2 + 0.5)

    # Define the range of p-values that will be possible to fit with a
    # polynomial function of order not larger than 2.
    min_fit_p, max_fit_p = p - 0.35, p + 0.35
    if min_fit_p < 0.5:
        min_fit_p = 0.5
    if max_fit_p > 0.985:
        max_fit_p = 0.985

    (n_sig, p_vals, p_val_weights) = ([], [], [])

    while True:
        ns_range_[0] = np.max([ns_range_[0], 0])
        logger.debug(
            'Doing new loop for nsignal range %s',
            str(ns_range_))

        ns0 = (ns_range_[1] + ns_range_[0]) / 2

        # Generate statistics (trials) for the current point ns0 as long as
        # the we are only 5sigma away from the desired propability and the
        # uncertainty of the probability is still larger than the desired
        # uncertainty ``eps_p``.
        # Initially generate trials for a 5-times larger uncertainty ``eps_p``
        # to catch ns0 points far away from the desired propability quicker.
        dn_trials = max(100, int(n_trials/5**2 + 0.5))
        (ts_vals0, p0_sigma, delta_p) = ([], 2*eps_p, 0)
        while (delta_p < p0_sigma*5) and (p0_sigma > eps_p):
            ts_vals0 = np.concatenate((
                ts_vals0, ana.do_trials(
                    rss=rss,
                    n=dn_trials,
                    mean_n_sig=ns0,
                    bkg_kwargs=bkg_kwargs,
                    sig_kwargs=sig_kwargs,
                    ppbar=ppbar,
                    tl=tl)['ts']))
            (p0, p0_sigma) = calculate_pval_from_trials(ts_vals0, c)

            n_total_generated_trials += dn_trials

            delta_p = np.abs(p0 - p)

            logger.debug(
                'n_trials: %d, ns0: %.6f, p0: %.6f, p0_sigma: %.6f, '
                'delta_p: %.6f',
                ts_vals0.size, ns0, p0, p0_sigma, delta_p)

            # After the initial number of trials generated the number of trials
            # to generate, dn_trials, for the next iteration of trial generation
            # to decrease p0_sigma can be set to the number of remaining trials
            # to reach n_trials. But do at least 100 trials more, in case the
            # n_trials estimate was initially too low.
            dn_trials = max(100, n_trials - ts_vals0.size)

            if (p0_sigma < eps_p) and (delta_p < eps_p):
                # We found the ns0 value that corresponds to the desired
                # probability within the desired uncertainty.

                if (p0 < max_fit_p) and (p0 > min_fit_p):
                    n_sig.append(ns0)
                    p_vals.append(p0)
                    p_val_weights.append(1. / p0_sigma)

                logger.debug(
                    'Found mu value %g with p value %g within uncertainty +-%g',
                    ns0, p0, p0_sigma)

                if p0 > p:
                    ns1 = ns_range_[0]
                    if np.abs(ns0 - ns1) > 1.0:
                        ns1 = ns0 - 1.0
                    if np.abs(ns0 - ns1) < min_dmu:
                        ns1 = ns0 - min_dmu
                else:
                    ns1 = ns_range_[1]
                    if np.abs(ns0 - ns1) > 1.0:
                        ns1 = ns0 + 1.0
                    if np.abs(ns0 - ns1) < min_dmu:
                        ns1 = ns0 + min_dmu

                ts_vals1 = ana.do_trials(
                    rss=rss,
                    n=ts_vals0.size,
                    mean_n_sig=ns1,
                    bkg_kwargs=bkg_kwargs,
                    sig_kwargs=sig_kwargs,
                    ppbar=ppbar,
                    tl=tl)['ts']
                n_total_generated_trials += ts_vals0.size

                (p1, p1_sigma) = calculate_pval_from_trials(ts_vals1, c)
                logger.debug(
                    'Final mu value is supposed to be within mu range (%g,%g) '
                    'corresponding to p=(%g +-%g, %g +-%g)',
                    ns0, ns1, p0, p0_sigma, p1, p1_sigma)

                if (p1 < max_fit_p) and (p1 > min_fit_p):
                    n_sig.append(ns1)
                    p_vals.append(p1)
                    p_val_weights.append(1. / p1_sigma)

                if len(n_sig) > 2:
                    scanned_range = np.max(n_sig) - np.min(n_sig)

                    if (len(n_sig) < 5) or (scanned_range < 1.5):
                        deg = 1
                    else:
                        deg = 2

                    logger.debug(
                        'Scanned mu range: [%g , %g]\nPoints to fit: %g\n '
                        'Using polynomial fit of order %g',
                        np.min(n_sig), np.max(n_sig), len(n_sig), deg)

                    mu = polynomial_fit(n_sig, p_vals, p_val_weights, deg, p)
                    mu_err = None

                else:
                    # If the points in the scanned range are only two, we
                    # calculate the final mu value with a linear interpolation
                    # between those points.

                    logger.debug(
                        'Scanned mu range: [%g , %g]\nPoints to fit: %g\n '
                        'Doing a linear interpolation.',
                        np.min(n_sig), np.max(n_sig), len(n_sig))

                    # Check if p1 and p0 are equal, which would result in a
                    # divison by zero.
                    if p0 == p1:
                        mu = 0.5*(ns0 + ns1)
                        mu_err = 0.5*np.abs(ns1 - ns0)

                        logger.debug(
                            'Probability for mu=%g and mu=%g has the same '
                            'value %g',
                            ns0, ns1, p0)
                    else:
                        dns_dp = np.abs((ns1 - ns0) / (p1 - p0))

                        logger.debug(
                            'Estimated |dmu/dp| = %g within mu range (%g,%g) '
                            'corresponding to p=(%g +-%g, %g +-%g)',
                            dns_dp, ns0, ns1, p0, p0_sigma, p1, p1_sigma)
                        if p0 > p:
                            mu = ns0 - dns_dp * delta_p
                        else:
                            mu = ns0 + dns_dp * delta_p
                        mu_err = dns_dp * delta_p

                logger.debug(
                    'Estimated final mu to be %g (error estimation to be '
                    'implemented)',
                    mu)

                return (mu, mu_err)

        if delta_p < p0_sigma*5:
            # The desired probability is within the 5 sigma region of the
            # current probability. So we use a linear approximation to find the
            # next ns range.
            # For the current ns0 the uncertainty of p0 is smaller than the
            # required uncertainty, hence p0_sigma <= eps_p.

            # Store ns0 for the new lower or upper bound depending on where the
            # p0 lies.

            if (p0 < max_fit_p) and (p0 > min_fit_p):
                n_sig.append(ns0)
                p_vals.append(p0)
                p_val_weights.append(1. / p0_sigma)

            if p0+p0_sigma+eps_p <= p:
                ns_lower_bound = ns0
            elif p0-p0_sigma-eps_p >= p:
                ns_upper_bound = ns0

            ns1 = ns0 * (1 - np.sign(p0 - p) * 0.05)
            if np.abs(ns0 - ns1) < min_dmu:
                if (p0 - p) < 0:
                    ns1 = ns0 + min_dmu
                else:
                    ns1 = ns0 - min_dmu

            logger.debug(
                'Do interpolation between ns=(%.3f, %.3f)',
                ns0, ns1)

            ts_vals1 = ana.do_trials(
                rss=rss,
                n=ts_vals0.size,
                mean_n_sig=ns1,
                bkg_kwargs=bkg_kwargs,
                sig_kwargs=sig_kwargs,
                ppbar=ppbar,
                tl=tl)['ts']
            n_total_generated_trials += ts_vals0.size

            (p1, p1_sigma) = calculate_pval_from_trials(ts_vals1, c)

            if (p1 < max_fit_p) and (p1 > min_fit_p):
                n_sig.append(ns1)
                p_vals.append(p1)
                p_val_weights.append(1. / p1_sigma)

            # Check if p0 and p1 are equal, which would result into a division
            # by zero.
            if p0 == p1:
                dp = 0.5*(p0_sigma + p1_sigma)
                logger.debug(
                    'p1 and p0 are equal to %g, causing division by zero. '
                    'p0_sigma=%g, p1_sigma=%g. Calculating dns/dp with dp=%g.',
                    p0, p0_sigma, p1_sigma, dp)
                dns_dp = np.abs((ns1 - ns0) / dp)
            else:
                dns_dp = np.abs((ns1 - ns0) / (p1 - p0))
                # p0 and p1 might be very similar, resulting into a numerically
                # infitite slope.
                if np.isinf(dns_dp):
                    dp = 0.5*(p0_sigma + p1_sigma)
                    logger.debug(
                        'Infinite dns/dp dedected: ns0=%g, ns1=%g, p0=%g, '
                        'p0_sigma=%g, p1=%g, p1_sigma=%g. Recalculating dns/dp '
                        'with deviation %g.',
                        ns0, ns1, p0, p0_sigma, p1, p1_sigma, dp)
                    dns_dp = np.abs((ns1 - ns0) / dp)
            logger.debug('dns/dp = %g', dns_dp)

            if p0 > p:
                ns_range_[0] = ns0 - dns_dp * (delta_p + p0_sigma)
                ns_range_[1] = ns0 + dns_dp * p0_sigma
            else:
                ns_range_[0] = ns0 - dns_dp * p0_sigma
                ns_range_[1] = ns0 + dns_dp * (delta_p + p0_sigma)

            # Restrict the range to ns values we already know well.
            ns_range_[0] = np.max((ns_range_[0], ns_lower_bound))
            ns_range_[1] = np.min((ns_range_[1], ns_upper_bound))

            # In case the new calculated mu range is smaller than the minimum
            # delta mu, the mu range gets widened by half of the minimum delta
            # mu on both sides.
            if np.abs(ns_range_[1] - ns_range_[0]) < min_dmu:
                ns_range_[0] -= 0.5*min_dmu
                ns_range_[1] += 0.5*min_dmu
        else:
            # The current ns corresponds to a probability p0 that is at least
            # 5 sigma away from the desired probability p, hence
            # delta_p >= p0_sigma*5.

            if (p0 < max_fit_p) and (p0 > min_fit_p):
                n_sig.append(ns0)
                p_vals.append(p0)
                p_val_weights.append(1. / p0_sigma)

            if p0 < p:
                ns_range_[0] = ns0
            else:
                ns_range_[1] = ns0

            if np.abs(ns_range_[1] - ns_range_[0]) < min_dmu:
                # The mu range became smaller than the minimum delta mu and
                # still beeing far away from the desired probability.
                # So move the mu range towards the desired probability.
                if p0 < p:
                    ns_range_[1] += 10*min_dmu
                else:
                    ns_range_[0] -= 10*min_dmu


def estimate_sensitivity(
        ana,
        rss,
        h0_trials=None,
        h0_ts_quantile=0.5,
        p=0.9,
        eps_p=0.005,
        mu_range=None,
        min_dmu=0.5,
        bkg_kwargs=None,
        sig_kwargs=None,
        ppbar=None,
        tl=None,
        pathfilename=None):
    """Estimates the mean number of signal events that whould have to be
    injected into the data such that the test-statistic value of p*100% of all
    trials are larger than the critical test-statistic value c, which
    corresponds to the test-statistic value where h0_ts_quantile*100% of all
    null hypothesis test-statistic values are larger than c.

    For sensitivity h0_ts_quantile, and p are usually set to 0.5, and 0.9,
    respectively.

    Parameters
    ----------
    ana : Analysis
        The Analysis instance to use for sensitivity estimation.
    rss : RandomStateService
        The RandomStateService instance to use for generating random
        numbers.
    h0_trials : (n_h0_ts_vals,)-shaped ndarray | None
        The strutured ndarray holding the trials for the null-hypothesis.
        If set to `None`, the number of trials is calculated from binomial
        statistics via `h0_ts_quantile*(1-h0_ts_quantile)/eps**2`,
        where `eps` is `min(5e-3, h0_ts_quantile/10)`.
    h0_ts_quantile : float, optional
        Null-hypothesis test statistic quantile that defines the critical value.
    p : float, optional
        Desired probability of the signal test statistic value to exceed
        the null-hypothesis test statistic value threshold, which is defined
        through the `h0_ts_quantile` value.
    eps_p : float, optional
        Precision in `p` for execution to break.
    mu_range : 2-element sequence | None
        Range to search for the mean number of signal events.
        If set to None, the range (0, 10) will be used.
    min_dmu : float
        The minimum delta mu to use for calculating the derivative dmu/dp.
        The default is ``0.5``.
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    sig_kwargs : dict | None
        Additional keyword arguments for the `generate_signal_events` method
        of the `SignalGenerator` class. An usual keyword argument is
        `poisson`. If `poisson` is set to True, the actual number of
        generated signal events will be drawn from a Poisson distribution
        with the mean number of signal events, mu.
    ppbar : instance of ProgressBar | None
        The possible parent ProgressBar instance.
    tl: instance of TimeLord | None
        The optional TimeLord instance that should be used to collect timing
        information about this function.
    pathfilename : string | None
        Trial data file path including the filename.
        If set to None, generated trials won't be saved.

    Returns
    -------
    mu : float
        Estimated median number of signal events to reach desired sensitivity.
    mu_err : float
        The uncertainty of the estimated mean number of signal events.
    """
    if mu_range is None:
        mu_range = (0, 10)

    (mu, mu_err) = estimate_mean_nsignal_for_ts_quantile(
        ana=ana,
        rss=rss,
        h0_trials=h0_trials,
        h0_ts_quantile=h0_ts_quantile,
        p=p,
        eps_p=eps_p,
        mu_range=mu_range,
        min_dmu=min_dmu,
        bkg_kwargs=bkg_kwargs,
        sig_kwargs=sig_kwargs,
        ppbar=ppbar,
        tl=tl,
        pathfilename=pathfilename)

    return (mu, mu_err)


def estimate_discovery_potential(
        ana,
        rss,
        h0_trials=None,
        h0_ts_quantile=2.8665e-7,
        p=0.5,
        eps_p=0.005,
        mu_range=None,
        min_dmu=0.5,
        bkg_kwargs=None,
        sig_kwargs=None,
        ppbar=None,
        tl=None,
        pathfilename=None):
    """Estimates the mean number of signal events that whould have to be
    injected into the data such that the test-statistic value of p*100% of all
    trials are larger than the critical test-statistic value c, which
    corresponds to the test-statistic value where h0_ts_quantile*100% of all
    null hypothesis test-statistic values are larger than c.

    For the 5 sigma discovery potential `h0_ts_quantile`, and `p` are usually
    set to 2.8665e-7, and 0.5, respectively.

    Parameters
    ----------
    ana : Analysis
        The Analysis instance to use for discovery potential estimation.
    rss : RandomStateService
        The RandomStateService instance to use for generating random
        numbers.
    h0_trials : (n_h0_ts_vals,)-shaped ndarray | None
        The structured ndarray holding the trials for the null-hypothesis.
        If set to `None`, the number of trials is calculated from binomial
        statistics via `h0_ts_quantile*(1-h0_ts_quantile)/eps**2`,
        where `eps` is `min(5e-3, h0_ts_quantile/10)`.
    h0_ts_quantile : float, optional
        Null-hypothesis test statistic quantile that defines the critical value.
    p : float, optional
        Desired probability of the signal test statistic value to exceed the
        critical value.
    eps_p : float, optional
        Precision in `p` for execution to break.
    mu_range : 2-element sequence | None
        Range to search for the mean number of signal events.
        If set to None, the range (0, 10) will be used.
    min_dmu : float
        The minimum delta mu to use for calculating the derivative dmu/dp.
        The default is ``0.5``.
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    sig_kwargs : dict | None
        Additional keyword arguments for the `generate_signal_events` method
        of the `SignalGenerator` class. An usual keyword argument is
        `poisson`. If `poisson` is set to True, the actual number of
        generated signal events will be drawn from a Poisson distribution
        with the mean number of signal events, mu.
    ppbar : instance of ProgressBar | None
        The possible parent ProgressBar instance.
    tl: instance of TimeLord | None
        The optional TimeLord instance that should be used to collect timing
        information about this function.
    pathfilename : string | None
        Trial data file path including the filename.
        If set to None, generated trials won't be saved.

    Returns
    -------
    mu : float
        Estimated mean number of injected signal events to reach the desired
        discovery potential.
    mu_err : float
        Estimated error of `mu`.
    """
    if mu_range is None:
        mu_range = (0, 10)

    (mu, mu_err) = estimate_mean_nsignal_for_ts_quantile(
        ana=ana,
        rss=rss,
        p=p,
        eps_p=eps_p,
        mu_range=mu_range,
        h0_trials=h0_trials,
        h0_ts_quantile=h0_ts_quantile,
        min_dmu=min_dmu,
        bkg_kwargs=bkg_kwargs,
        sig_kwargs=sig_kwargs,
        ppbar=ppbar,
        tl=tl,
        pathfilename=pathfilename)

    return (mu, mu_err)


def generate_mu_of_p_spline_interpolation(
        ana,
        rss,
        h0_ts_vals,
        h0_ts_quantile,
        eps_p,
        mu_range,
        mu_step,
        kind='cubic',
        bkg_kwargs=None,
        sig_kwargs=None,
        ppbar=None,
        tl=None):
    """Generates a spline interpolation for mu(p) function for a pre-defined
    range of mu, where mu is the mean number of injected signal events and p the
    probability for the ts value larger than the ts value corresponding to the
    given quantile, h0_ts_quantile, of the null hypothesis test-statistic
    distribution.

    Parameters
    ----------
    ana : instance of Analysis
        The Analysis instance to use for the calculation.
    rss : instance of RandomStateService
        The RandomStateService instance to use for generating random numbers.
    h0_ts_vals : (n_h0_ts_vals,)-shaped 1D ndarray | None
        The 1D ndarray holding the test-statistic values for the
        null-hypothesis. If set to `None`, 100/(1-h0_ts_quantile)
        null-hypothesis trials will be generated.
    h0_ts_quantile : float
        Null-hypothesis test statistic quantile, which should be exceeded by
        the alternative hypothesis ts value.
    eps_p : float
        The one sigma precision in `p` as stopping condition for the
        calculation for a single mu value.
    mu_range : 2-element sequence
        The range (lower,upper) of mean number of injected signal events to
        create the interpolation spline for.
    mu_step : float
        The step size of the mean number of signal events.
    kind : str
        The kind of spline to generate. Possble values are 'linear' and 'cubic'
        (default).
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    sig_kwargs : dict | None
        Additional keyword arguments for the `generate_signal_events` method
        of the `SignalGenerator` class. An usual keyword argument is
        `poisson`. If `poisson` is set to True, the actual number of
        generated signal events will be drawn from a Poisson distribution
        with the mean number of signal events, mu.
    ppbar : instance of ProgressBar | None
        The possible parent ProgressBar instance.
    tl: instance of TimeLord | None
        The optional TimeLord instance that should be used to collect timing
        information about this function.

    Returns
    -------
    spline : callable
        The spline function mu(p).
    """
    logger = logging.getLogger(__name__)

    n_total_generated_trials = 0

    if h0_ts_vals is None:
        n_bkg = int(100/(1 - h0_ts_quantile))
        logger.debug('Generate %d null-hypothesis trials', n_bkg)
        h0_ts_vals = ana.do_trials(
            rss=rss,
            n=n_bkg,
            mean_n_sig=0,
            bkg_kwargs=bkg_kwargs,
            sig_kwargs=sig_kwargs,
            ppbar=ppbar,
            tl=tl)['ts']
        n_total_generated_trials += n_bkg

    n_h0_ts_vals = len(h0_ts_vals)
    h0_ts_vals = h0_ts_vals[np.isfinite(h0_ts_vals)]
    logger.debug(
        'Number of trials after finite cut: %d (%g%% of total)',
        len(h0_ts_vals), (len(h0_ts_vals)/n_h0_ts_vals)*100)

    c = np.percentile(h0_ts_vals, (1 - h0_ts_quantile) * 100)
    logger.debug(
        'Critical ts value for bkg ts quantile %g: %g',
        h0_ts_quantile, c)

    n_mu = int((mu_range[1]-mu_range[0])/mu_step) + 1
    mu_vals = np.linspace(mu_range[0], mu_range[1], n_mu)
    p_vals = np.empty_like(mu_vals)

    logger.debug(
        'Generate trials for %d mu values',
        n_mu)

    # Create the progress bar if we are in an interactive session.
    pbar = None
    if is_interactive_session():
        pbar = ProgressBar(len(mu_vals), parent=ppbar).start()

    for (idx, mu) in enumerate(mu_vals):
        p = None
        (ts_vals, p_sigma) = ([], 2*eps_p)
        while (p_sigma > eps_p):
            ts_vals = np.concatenate(
                (ts_vals,
                 ana.do_trials(
                     rss=rss,
                     n=100,
                     mean_n_sig=mu,
                     bkg_kwargs=bkg_kwargs,
                     sig_kwargs=sig_kwargs,
                     ppbar=pbar,
                     tl=tl)['ts']))
            (p, p_sigma) = calculate_pval_from_trials(ts_vals, c)
            n_total_generated_trials += 100
        logger.debug(
            'mu: %g, n_trials: %d, p: %g, p_sigma: %g',
            mu, ts_vals.size, p, p_sigma)
        p_vals[idx] = p

        if pbar is not None:
            pbar.increment()

    # Make a mu(p) spline via interp1d.
    spline = make_spline_1d(
        p_vals,
        mu_vals,
        kind=kind,
        copy=False,
        assume_sorted=True)

    if pbar is not None:
        pbar.finish()

    return spline


def create_trial_data_file(  # noqa: C901
        ana,
        rss,
        n_trials,
        mean_n_sig=0,
        mean_n_sig_null=0,
        mean_n_bkg_list=None,
        bkg_kwargs=None,
        sig_kwargs=None,
        pathfilename=None,
        ncpu=None,
        ppbar=None,
        tl=None):
    """Creates and fills a trial data file with `n_trials` generated trials for
    each mean number of injected signal events from `ns_min` up to `ns_max` for
    a given analysis.

    Parameters
    ----------
    ana : instance of Analysis
        The Analysis instance to use for the trial generation.
    rss : instance of RandomStateService
        The RandomStateService instance to use for generating random
        numbers.
    n_trials : int
        The number of trials to perform for each hypothesis test.
    mean_n_sig : ndarray of float | float | 2- or 3-element sequence of float
        The array of mean number of injected signal events (MNOISEs) for which
        to generate trials. If this argument is not a ndarray, an array of
        MNOISEs is generated based on this argument.
        If a single float is given, only this given MNOISEs are injected.
        If a 2-element sequence of floats is given, it specifies the range of
        MNOISEs with a step size of one.
        If a 3-element sequence of floats is given, it specifies the range plus
        the step size of the MNOISEs.
    mean_n_sig_null : ndarray of float | float | 2- or 3-element sequence of float
        The array of the fixed mean number of signal events (FMNOSEs) for the
        null-hypothesis for which to generate trials. If this argument is not a
        ndarray, an array of FMNOSEs is generated based on this argument.
        If a single float is given, only this given FMNOSEs are used.
        If a 2-element sequence of floats is given, it specifies the range of
        FMNOSEs with a step size of one.
        If a 3-element sequence of floats is given, it specifies the range plus
        the step size of the FMNOSEs.
    mean_n_bkg_list : list of float | None
        The mean number of background events that should be generated for
        each dataset. This parameter is passed to the ``do_trials`` method of
        the ``Analysis`` class. If set to None (the default), the background
        generation method needs to obtain this number itself.
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    sig_kwargs : dict | None
        Additional keyword arguments for the `generate_signal_events` method
        of the `SignalGenerator` class. An usual keyword argument is
        `poisson`.
    pathfilename : string | None
        Trial data file path including the filename.
        If set to None generated trials won't be saved.
    ncpu : int | None
        The number of CPUs to use.
    ppbar : instance of ProgressBar | None
        The optional instance of the parent progress bar.
    tl: instance of TimeLord | None
        The instance of TimeLord that should be used to measure individual
        tasks.

    Returns
    -------
    seed : int
        The seed used to generate the trials.
    mean_n_sig : 1d ndarray
        The array holding the mean number of signal events used to generate the
        trials.
    mean_n_sig_null : 1d ndarray
        The array holding the fixed mean number of signal events for the
        null-hypothesis used to generate the trials.
    trial_data : structured numpy ndarray
        The generated trial data.
    """
    n_trials = int_cast(
        n_trials,
        'The n_trials argument must be castable to type int!')

    if not isinstance(mean_n_sig, np.ndarray):
        if not issequence(mean_n_sig):
            mean_n_sig = float_cast(
                mean_n_sig,
                'The mean_n_sig argument must be castable to type float!')
            mean_n_sig_min = mean_n_sig
            mean_n_sig_max = mean_n_sig
            mean_n_sig_step = 1
        else:
            mean_n_sig = float_cast(
                mean_n_sig,
                'The sequence elements of the mean_n_sig argument must be '
                'castable to float values!')
            if len(mean_n_sig) == 2:
                (mean_n_sig_min, mean_n_sig_max) = mean_n_sig
                mean_n_sig_step = 1
            elif len(mean_n_sig) == 3:
                (mean_n_sig_min, mean_n_sig_max, mean_n_sig_step) = mean_n_sig

        mean_n_sig = np.arange(
            mean_n_sig_min, mean_n_sig_max+1, mean_n_sig_step,
            dtype=np.float64)

    if not isinstance(mean_n_sig_null, np.ndarray):
        if not issequence(mean_n_sig_null):
            mean_n_sig_null = float_cast(
                mean_n_sig_null,
                'The mean_n_sig_null argument must be castable to type float!')
            mean_n_sig_null_min = mean_n_sig_null
            mean_n_sig_null_max = mean_n_sig_null
            mean_n_sig_null_step = 1
        else:
            mean_n_sig_null = float_cast(
                mean_n_sig_null,
                'The sequence elements of the mean_n_sig_null argument must '
                'be castable to float values!')
            if len(mean_n_sig_null) == 2:
                (mean_n_sig_null_min, mean_n_sig_null_max) = mean_n_sig_null
                mean_n_sig_null_step = 1
            elif len(mean_n_sig_null) == 3:
                (mean_n_sig_null_min, mean_n_sig_null_max,
                 mean_n_sig_null_step) = mean_n_sig_null

        mean_n_sig_null = np.arange(
            mean_n_sig_null_min, mean_n_sig_null_max+1, mean_n_sig_null_step,
            dtype=np.float64)

    pbar = ProgressBar(
        len(mean_n_sig)*len(mean_n_sig_null), parent=ppbar).start()
    trial_data = None
    for (mean_n_sig_, mean_n_sig_null_) in itertools.product(
            mean_n_sig, mean_n_sig_null):

        trials = ana.do_trials(
            rss=rss,
            n=n_trials,
            mean_n_bkg_list=mean_n_bkg_list,
            mean_n_sig=mean_n_sig_,
            mean_n_sig_0=mean_n_sig_null_,
            bkg_kwargs=bkg_kwargs,
            sig_kwargs=sig_kwargs,
            ncpu=ncpu,
            tl=tl,
            ppbar=pbar)

        if trial_data is None:
            trial_data = trials
        else:
            trial_data = np_rfn.stack_arrays(
                [trial_data, trials],
                usemask=False,
                asrecarray=True)

        pbar.increment()
    pbar.finish()

    if trial_data is None:
        raise RuntimeError(
            'No trials have been generated! Check your generation boundaries!')

    if pathfilename is not None:
        dirname = os.path.dirname(pathfilename)
        if dirname:
            # Create the directory if dirname is not empty.
            makedirs(dirname, exist_ok=True)
        # Save the trial data to file.
        np.save(pathfilename, trial_data)

    return (rss.seed, mean_n_sig, mean_n_sig_null, trial_data)


def extend_trial_data_file(
        ana,
        rss,
        n_trials,
        trial_data,
        mean_n_sig=0,
        mean_n_sig_null=0,
        mean_n_bkg_list=None,
        bkg_kwargs=None,
        sig_kwargs=None,
        pathfilename=None,
        **kwargs):
    """Appends to the trial data file `n_trials` generated trials for each
    mean number of injected signal events up to `ns_max` for a given analysis.

    Parameters
    ----------
    ana : instance of Analysis
        The Analysis instance to use for sensitivity estimation.
    rss : instance of RandomStateService
        The RandomStateService instance to use for generating random
        numbers.
    n_trials : int
        The number of trials the trial data file needs to be extended by.
    trial_data : structured numpy ndarray
        The structured numpy ndarray holding the trials.
    mean_n_sig : ndarray of float | float | 2- or 3-element sequence of float
        The array of mean number of injected signal events (MNOISEs) for which
        to generate trials. If this argument is not a ndarray, an array of
        MNOISEs is generated based on this argument.
        If a single float is given, only this given MNOISEs are injected.
        If a 2-element sequence of floats is given, it specifies the range of
        MNOISEs with a step size of one.
        If a 3-element sequence of floats is given, it specifies the range plus
        the step size of the MNOISEs.
    mean_n_sig_null : ndarray of float | float | 2- or 3-element sequence of float
        The array of the fixed mean number of signal events (FMNOSEs) for the
        null-hypothesis for which to generate trials. If this argument is not a
        ndarray, an array of FMNOSEs is generated based on this argument.
        If a single float is given, only this given FMNOSEs are used.
        If a 2-element sequence of floats is given, it specifies the range of
        FMNOSEs with a step size of one.
        If a 3-element sequence of floats is given, it specifies the range plus
        the step size of the FMNOSEs.
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    sig_kwargs : dict | None
        Additional keyword arguments for the `generate_signal_events` method
        of the `SignalGenerator` class. An usual keyword argument is
        `poisson`.
    pathfilename : string | None
        Trial data file path including the filename.

    Additional keyword arguments
    ----------------------------
    Additional keyword arguments are passed-on to the ``create_trial_data_file``
    function.

    Returns
    -------
    trial_data :
        Trial data file extended by the required number of trials for each
        mean number of injected signal events..
    """
    # Use unique seed to generate non identical trials.
    if rss.seed in trial_data['seed']:
        seed = next(
            i
            for (i, e) in enumerate(
                sorted(np.unique(trial_data['seed'])) + [None], 1)
            if i != e)
        rss.reseed(seed)

    (seed, mean_n_sig, mean_n_sig_null, trials) = create_trial_data_file(
        ana=ana,
        rss=rss,
        n_trials=n_trials,
        mean_n_sig=mean_n_sig,
        mean_n_sig_null=mean_n_sig_null,
        mean_n_bkg_list=mean_n_bkg_list,
        bkg_kwargs=bkg_kwargs,
        sig_kwargs=sig_kwargs,
        **kwargs
    )
    trial_data = np_rfn.stack_arrays(
        [trial_data, trials],
        usemask=False,
        asrecarray=True)

    if pathfilename is not None:
        dirname = os.path.dirname(pathfilename)
        if dirname:
            # Create the directory if dirname is not empty.
            makedirs(dirname, exist_ok=True)
        # Save the trial data to file.
        np.save(pathfilename, trial_data)

    return trial_data


def calculate_upper_limit_distribution(
        ana,
        rss,
        pathfilename,
        n_bkg=5000,
        n_bins=100):
    """Function to calculate upper limit distribution. It loads the trial data
    file containing test statistic distribution and calculates 10 percentile
    value for each mean number of injected signal event. Then it finds upper
    limit values which correspond to generated background trials test statistic
    values by linearly interpolated curve of 10 percentile values distribution.

    Parameters
    ----------
    ana : instance of Analysis
        The Analysis instance to use for sensitivity estimation.
    rss : instance of RandomStateService
        The RandomStateService instance to use for generating random
        numbers.
    pathfilename : string
        Trial data file path including the filename.
    n_bkg : int, optional
        Number of times to perform background analysis trial.
    n_bins : int, optional
        Number of returned test statistic histograms bins.

    Returns
    -------
    result : dict
        Result dictionary which contains the following fields:

        ul : list of float
            List of upper limit values.
        mean : float
            Mean of upper limit values.
        median : float
            Median of upper limit values.
        var : float
            Variance of upper limit values.
        ts_hist : numpy ndarray
            2D array of test statistic histograms calculated by axis 1.
        extent : list of float
            Test statistic histogram boundaries.
        q_values : list of float
            `q` percentile values of test statistic for different injected
            events means.
    """
    # Load trial data file.
    trial_data = NPYFileLoader(pathfilename).load_data()
    ns_max = max(trial_data['sig_mean']) + 1
    ts_bins_range = (min(trial_data['TS']), max(trial_data['TS']))

    q = 10  # Upper limit criterion.
    trial_data_q_values = np.empty((ns_max,))
    trial_data_ts_hist = np.empty((ns_max, n_bins))
    for ns in range(ns_max):
        trial_data_q_values[ns] = np.percentile(
            trial_data['TS'][trial_data['sig_mean'] == ns], q)
        (trial_data_ts_hist[ns, :], bin_edges) = np.histogram(
            trial_data['TS'][trial_data['sig_mean'] == ns],
            bins=n_bins, range=ts_bins_range)

    ts_inv_f = interp1d(trial_data_q_values, range(ns_max), kind='linear')
    ts_bkg = ana.do_trials(
        rss=rss,
        n=n_bkg,
        mean_n_sig=0)['TS']

    # Cut away lower background test statistic values than the minimal
    # `ts_inv_f` interpolation boundary.
    ts_bkg = ts_bkg[ts_bkg >= min(trial_data_q_values)]

    ul_list = map(ts_inv_f, ts_bkg)
    ul_mean = np.mean(ul_list)
    ul_median = np.median(ul_list)
    ul_var = np.var(ul_list)

    result = {}
    result['ul'] = ul_list
    result['mean'] = ul_mean
    result['median'] = ul_median
    result['var'] = ul_var
    result['ts_hist'] = trial_data_ts_hist
    result['extent'] = [0, ns_max, ts_bins_range[0], ts_bins_range[1]]
    result['q_values'] = trial_data_q_values

    return result



# -*- coding: utf-8 -*-

from skyllh.core import (
    session,
)

# Automatically enable interactive mode, if the Python interpreter is in
# interactive mode.
if session.is_python_interpreter_in_interactive_mode():
    session.enable_interactive_session()
else:
    session.disable_interactive_session()



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.storage import create_FileLoader


def load_smearing_histogram(
        pathfilenames):
    """Loads the 5D smearing histogram from the given data file.

    Parameters
    ----------
    pathfilenames : str | list of str
        The file name of the data file.

    Returns
    -------
    histogram : 5d ndarray
        The 5d histogram array holding the probability values of the smearing
        matrix.
        The axes are (true_e, true_dec, reco_e, psi, ang_err).
    true_e_bin_edges : 1d ndarray
        The ndarray holding the bin edges of the true energy axis.
    true_dec_bin_edges : 1d ndarray
        The ndarray holding the bin edges of the true declination axis in
        radians.
    reco_e_lower_edges : 3d ndarray
        The 3d ndarray holding the lower bin edges of the reco energy axis.
        For each pair of true_e and true_dec different reco energy bin edges
        are provided.
        The shape is (n_true_e, n_true_dec, n_reco_e).
    reco_e_upper_edges : 3d ndarray
        The 3d ndarray holding the upper bin edges of the reco energy axis.
        For each pair of true_e and true_dec different reco energy bin edges
        are provided.
        The shape is (n_true_e, n_true_dec, n_reco_e).
    psi_lower_edges : 4d ndarray
        The 4d ndarray holding the lower bin edges of the psi axis in radians.
        The shape is (n_true_e, n_true_dec, n_reco_e, n_psi).
    psi_upper_edges : 4d ndarray
        The 4d ndarray holding the upper bin edges of the psi axis in radians.
        The shape is (n_true_e, n_true_dec, n_reco_e, n_psi).
    ang_err_lower_edges : 5d ndarray
        The 5d ndarray holding the lower bin edges of the angular error axis
        in radians.
        The shape is (n_true_e, n_true_dec, n_reco_e, n_psi, n_ang_err).
    ang_err_upper_edges : 5d ndarray
        The 5d ndarray holding the upper bin edges of the angular error axis
        in radians.
        The shape is (n_true_e, n_true_dec, n_reco_e, n_psi, n_ang_err).
    """
    # Load the smearing data from the public dataset.
    loader = create_FileLoader(pathfilenames=pathfilenames)
    data = loader.load_data()
    # Rename the data fields.
    renaming_dict = {
        'log10(E_nu/GeV)_min': 'true_e_min',
        'log10(E_nu/GeV)_max': 'true_e_max',
        'Dec_nu_min[deg]':     'true_dec_min',
        'Dec_nu_max[deg]':     'true_dec_max',
        'log10(E/GeV)_min':    'e_min',
        'log10(E/GeV)_max':    'e_max',
        'PSF_min[deg]':        'psi_min',
        'PSF_max[deg]':        'psi_max',
        'AngErr_min[deg]':     'ang_err_min',
        'AngErr_max[deg]':     'ang_err_max',
        'Fractional_Counts':   'norm_counts'
    }
    data.rename_fields(renaming_dict)

    def _get_nbins_from_edges(lower_edges, upper_edges):
        """Helper function to extract the number of bins from the data's
        bin edges.
        """
        n = 0
        # Select only valid rows.
        mask = (upper_edges - lower_edges) > 0
        data = lower_edges[mask]
        # Go through the valid rows and search for the number of increasing
        # bin edge values.
        v0 = None
        for v in data:
            if (v0 is not None) and (v < v0):
                # Reached the end of the edges block.
                break
            if (v0 is None) or (v > v0):
                v0 = v
                n += 1
        return n

    true_e_bin_edges = np.union1d(
        data['true_e_min'], data['true_e_max'])
    true_dec_bin_edges = np.union1d(
        data['true_dec_min'], data['true_dec_max'])

    n_true_e = len(true_e_bin_edges) - 1
    n_true_dec = len(true_dec_bin_edges) - 1

    n_reco_e = _get_nbins_from_edges(
        data['e_min'], data['e_max'])
    n_psi = _get_nbins_from_edges(
        data['psi_min'], data['psi_max'])
    n_ang_err = _get_nbins_from_edges(
        data['ang_err_min'], data['ang_err_max'])

    # Get reco energy bin_edges as a 3d array.
    idxs = np.array(
        range(len(data))
    ) % (n_psi * n_ang_err) == 0

    reco_e_lower_edges = np.reshape(
        data['e_min'][idxs],
        (n_true_e, n_true_dec, n_reco_e)
    )
    reco_e_upper_edges = np.reshape(
        data['e_max'][idxs],
        (n_true_e, n_true_dec, n_reco_e)
    )

    # Get psi bin_edges as a 4d array.
    idxs = np.array(
        range(len(data))
    ) % n_ang_err == 0

    psi_lower_edges = np.reshape(
        data['psi_min'][idxs],
        (n_true_e, n_true_dec, n_reco_e, n_psi)
    )
    psi_upper_edges = np.reshape(
        data['psi_max'][idxs],
        (n_true_e, n_true_dec, n_reco_e, n_psi)
    )

    # Get angular error bin_edges as a 5d array.
    ang_err_lower_edges = np.reshape(
        data['ang_err_min'],
        (n_true_e, n_true_dec, n_reco_e, n_psi, n_ang_err)
    )
    ang_err_upper_edges = np.reshape(
        data['ang_err_max'],
        (n_true_e, n_true_dec, n_reco_e, n_psi, n_ang_err)
    )

    # Create 5D histogram for the probabilities.
    histogram = np.reshape(
        data['norm_counts'],
        (
            n_true_e,
            n_true_dec,
            n_reco_e,
            n_psi,
            n_ang_err
        )
    )

    # Convert degrees into radians.
    true_dec_bin_edges = np.radians(true_dec_bin_edges)
    psi_lower_edges = np.radians(psi_lower_edges)
    psi_upper_edges = np.radians(psi_upper_edges)
    ang_err_lower_edges = np.radians(ang_err_lower_edges)
    ang_err_upper_edges = np.radians(ang_err_upper_edges)

    return (
        histogram,
        true_e_bin_edges,
        true_dec_bin_edges,
        reco_e_lower_edges,
        reco_e_upper_edges,
        psi_lower_edges,
        psi_upper_edges,
        ang_err_lower_edges,
        ang_err_upper_edges
    )


class PDSmearingMatrix(
        object):
    """This class is a helper class for dealing with the smearing matrix
    provided by the public data.
    """
    def __init__(
            self,
            pathfilenames,
            **kwargs):
        """Creates a smearing matrix instance by loading the smearing matrix
        from the given file.
        """
        super().__init__(**kwargs)

        (
            self.histogram,
            self._true_e_bin_edges,
            self._true_dec_bin_edges,
            self.reco_e_lower_edges,
            self.reco_e_upper_edges,
            self.psi_lower_edges,
            self.psi_upper_edges,
            self.ang_err_lower_edges,
            self.ang_err_upper_edges
        ) = load_smearing_histogram(pathfilenames)

        self.n_psi_bins = self.histogram.shape[3]
        self.n_ang_err_bins = self.histogram.shape[4]

        # Create bin edges array for log10_reco_e.
        s = np.array(self.reco_e_lower_edges.shape)
        s[-1] += 1
        self.log10_reco_e_binedges = np.empty(s, dtype=np.double)
        self.log10_reco_e_binedges[..., :-1] = self.reco_e_lower_edges
        self.log10_reco_e_binedges[..., -1] = self.reco_e_upper_edges[..., -1]

        # Create bin edges array for psi.
        s = np.array(self.psi_lower_edges.shape)
        s[-1] += 1
        self.psi_binedges = np.empty(s, dtype=np.double)
        self.psi_binedges[..., :-1] = self.psi_lower_edges
        self.psi_binedges[..., -1] = self.psi_upper_edges[..., -1]

        # Create bin edges array for ang_err.
        s = np.array(self.ang_err_lower_edges.shape)
        s[-1] += 1
        self.ang_err_binedges = np.empty(s, dtype=np.double)
        self.ang_err_binedges[..., :-1] = self.ang_err_lower_edges
        self.ang_err_binedges[..., -1] = self.ang_err_upper_edges[..., -1]

    @property
    def n_log10_true_e_bins(self):
        """(read-only) The number of log10 true energy bins.
        """
        return len(self._true_e_bin_edges) - 1

    @property
    def true_e_bin_edges(self):
        """(read-only) The (n_true_e+1,)-shaped 1D numpy ndarray holding the
        bin edges of the true energy.

        Depricated! Use log10_true_enu_binedges instead!
        """
        return self._true_e_bin_edges

    @property
    def true_e_bin_centers(self):
        """(read-only) The (n_true_e,)-shaped 1D numpy ndarray holding the bin
        center values of the true energy.
        """
        return 0.5*(self._true_e_bin_edges[:-1] +
                    self._true_e_bin_edges[1:])

    @property
    def log10_true_enu_binedges(self):
        """(read-only) The (n_log10_true_enu+1,)-shaped 1D numpy ndarray holding
        the bin edges of the log10 true neutrino energy.
        """
        return self._true_e_bin_edges

    @property
    def n_true_dec_bins(self):
        """(read-only) The number of true declination bins.
        """
        return len(self._true_dec_bin_edges) - 1

    @property
    def true_dec_bin_edges(self):
        """(read-only) The (n_true_dec+1,)-shaped 1D numpy ndarray holding the
        bin edges of the true declination.
        """
        return self._true_dec_bin_edges

    @property
    def true_dec_bin_centers(self):
        """(read-only) The (n_true_dec,)-shaped 1D ndarray holding the bin
        center values of the true declination.
        """
        return 0.5*(self._true_dec_bin_edges[:-1] +
                    self._true_dec_bin_edges[1:])

    @property
    def log10_reco_e_binedges_lower(self):
        """(read-only) The upper bin edges of the log10 reco energy axes.
        """
        return self.reco_e_lower_edges

    @property
    def log10_reco_e_binedges_upper(self):
        """(read-only) The upper bin edges of the log10 reco energy axes.
        """
        return self.reco_e_upper_edges

    @property
    def min_log10_reco_e(self):
        """(read-only) The minimum value of the reconstructed energy axis.
        """
        # Select only valid reco energy bins with bin widths greater than zero.
        m = (self.reco_e_upper_edges - self.reco_e_lower_edges) > 0
        return np.min(self.reco_e_lower_edges[m])

    @property
    def max_log10_reco_e(self):
        """(read-only) The maximum value of the reconstructed energy axis.
        """
        # Select only valid reco energy bins with bin widths greater than zero.
        m = (self.reco_e_upper_edges - self.reco_e_lower_edges) > 0
        return np.max(self.reco_e_upper_edges[m])

    @property
    def min_log10_psi(self):
        """(read-only) The minimum log10 value of the psi axis.
        """
        # Select only valid psi bins with bin widths greater than zero.
        m = (self.psi_upper_edges - self.psi_lower_edges) > 0
        return np.min(np.log10(self.psi_lower_edges[m]))

    @property
    def max_log10_psi(self):
        """(read-only) The maximum log10 value of the psi axis.
        """
        # Select only valid psi bins with bin widths greater than zero.
        m = (self.psi_upper_edges - self.psi_lower_edges) > 0
        return np.max(np.log10(self.psi_upper_edges[m]))

    @property
    def pdf(self):
        """(read-only) The probability-density-function
        P(E_reco,psi,ang_err|E_nu,dec_nu), which, by definition, is the
        histogram property divided by the 3D bin volumes for E_reco, psi, and
        ang_err.
        """
        log10_reco_e_bw = self.reco_e_upper_edges - self.reco_e_lower_edges
        psi_bw = self.psi_upper_edges - self.psi_lower_edges
        ang_err_bw = self.ang_err_upper_edges - self.ang_err_lower_edges

        bin_volumes = (
            log10_reco_e_bw[
                :, :, :, np.newaxis, np.newaxis
            ] *
            psi_bw[
                :, :, :, :, np.newaxis
            ] *
            ang_err_bw[
                :, :, :, :, :
            ]
        )

        # Divide the histogram bin probability values by their bin volume.
        # We do this only where the histogram actually has non-zero entries.
        pdf = np.copy(self.histogram)
        m = self.histogram != 0
        pdf[m] /= bin_volumes[m]

        return pdf

    def get_true_dec_idx(
            self,
            true_dec):
        """Returns the true declination index for the given true declination
        value.

        Parameters
        ----------
        true_dec : float
            The true declination value in radians.

        Returns
        -------
        true_dec_idx : int
            The index of the declination bin for the given declination value.
        """
        if (true_dec < self.true_dec_bin_edges[0]) or\
           (true_dec > self.true_dec_bin_edges[-1]):
            raise ValueError(
                f'The declination {true_dec} degrees is not supported by the '
                'smearing matrix!')

        true_dec_idx = np.digitize(true_dec, self.true_dec_bin_edges) - 1

        return true_dec_idx

    def get_log10_true_e_idx(
            self,
            log10_true_e):
        """Returns the bin index for the given true log10 energy value.

        Parameters
        ----------
        log10_true_e : float
            The log10 value of the true energy.

        Returns
        -------
        log10_true_e_idx : int
            The index of the true log10 energy bin for the given log10 true
            energy value.
        """
        if (log10_true_e < self.true_e_bin_edges[0]) or\
           (log10_true_e > self.true_e_bin_edges[-1]):
            raise ValueError(
                f'The log10 true energy value {log10_true_e} is not supported '
                'by the smearing matrix!')

        log10_true_e_idx = np.digitize(
            log10_true_e, self._true_e_bin_edges) - 1

        return log10_true_e_idx

    def get_reco_e_idx(
            self,
            true_e_idx,
            true_dec_idx,
            reco_e):
        """Returns the bin index for the given reco energy value given the
        given true energy and true declination bin indices.

        Parameters
        ----------
        true_e_idx : int
            The index of the true energy bin.
        true_dec_idx : int
            The index of the true declination bin.
        reco_e : float
            The reco energy value for which the bin index should get returned.

        Returns
        -------
        reco_e_idx : int | None
            The index of the reco energy bin the given reco energy value falls
            into. It returns None if the value is out of range.
        """
        lower_edges = self.reco_e_lower_edges[true_e_idx, true_dec_idx]
        upper_edges = self.reco_e_upper_edges[true_e_idx, true_dec_idx]

        m = (lower_edges <= reco_e) & (upper_edges > reco_e)
        idxs = np.nonzero(m)[0]
        if (len(idxs) == 0):
            return None

        reco_e_idx = idxs[0]

        return reco_e_idx

    def get_psi_idx(
            self,
            true_e_idx,
            true_dec_idx,
            reco_e_idx,
            psi):
        """Returns the bin index for the given psi value given the
        true energy, true declination and reco energy bin indices.

        Parameters
        ----------
        true_e_idx : int
            The index of the true energy bin.
        true_dec_idx : int
            The index of the true declination bin.
        reco_e_idx : int
            The index of the reco energy bin.
        psi : float
            The psi value in radians for which the bin index should get
            returned.

        Returns
        -------
        psi_idx : int | None
            The index of the psi bin the given psi value falls into.
            It returns None if the value is out of range.
        """
        lower_edges = self.psi_lower_edges[true_e_idx, true_dec_idx, reco_e_idx]
        upper_edges = self.psi_upper_edges[true_e_idx, true_dec_idx, reco_e_idx]

        m = (lower_edges <= psi) & (upper_edges > psi)
        idxs = np.nonzero(m)[0]
        if len(idxs) == 0:
            return None

        psi_idx = idxs[0]

        return psi_idx

    def get_ang_err_idx(
            self,
            true_e_idx,
            true_dec_idx,
            reco_e_idx,
            psi_idx,
            ang_err):
        """Returns the bin index for the given angular error value given the
        true energy, true declination, reco energy, and psi bin indices.

        Parameters
        ----------
        true_e_idx : int
            The index of the true energy bin.
        true_dec_idx : int
            The index of the true declination bin.
        reco_e_idx : int
            The index of the reco energy bin.
        psi_idx : int
            The index of the psi bin.
        ang_err : float
            The angular error value in radians for which the bin index should
            get returned.

        Returns
        -------
        ang_err_idx : int | None
            The index of the angular error bin the given angular error value
            falls into. It returns None if the value is out of range.
        """
        lower_edges = self.ang_err_lower_edges[
            true_e_idx, true_dec_idx, reco_e_idx, psi_idx]
        upper_edges = self.ang_err_upper_edges[
            true_e_idx, true_dec_idx, reco_e_idx, psi_idx]

        m = (lower_edges <= ang_err) & (upper_edges > ang_err)
        idxs = np.nonzero(m)[0]
        if len(idxs) == 0:
            return None

        ang_err_idx = idxs[0]

        return ang_err_idx

    def get_true_log_e_range_with_valid_log_e_pdfs(
            self,
            dec_idx):
        """Determines the true log energy range for which log_e PDFs are
        available for the given declination bin.

        Parameters
        ----------
        dec_idx : int
            The declination bin index.

        Returns
        -------
        min_log_true_e : float
            The minimum true log energy value.
        max_log_true_e : float
            The maximum true log energy value.
        """
        m = np.sum(
            (self.reco_e_upper_edges[:, dec_idx] -
             self.reco_e_lower_edges[:, dec_idx] > 0),
            axis=1) != 0
        min_log_true_e = np.min(self.true_e_bin_edges[:-1][m])
        max_log_true_e = np.max(self.true_e_bin_edges[1:][m])

        return (min_log_true_e, max_log_true_e)

    def get_log_e_pdf(
            self,
            log_true_e_idx,
            dec_idx):
        """Retrieves the log_e PDF from the given true energy bin index and
        source bin index.
        Returns (None, None, None, None) if any of the bin indices are less then
        zero, or if the sum of all pdf bins is zero.

        Parameters
        ----------
        log_true_e_idx : int
            The index of the true energy bin.
        dec_idx : int
            The index of the declination bin.

        Returns
        -------
        pdf : 1d ndarray
            The log_e pdf values.
        lower_bin_edges : 1d ndarray
            The lower bin edges of the energy pdf histogram.
        upper_bin_edges : 1d ndarray
            The upper bin edges of the energy pdf histogram.
        bin_widths : 1d ndarray
            The bin widths of the energy pdf histogram.
        """
        if log_true_e_idx < 0 or dec_idx < 0:
            return (None, None, None, None)

        pdf = self.histogram[log_true_e_idx, dec_idx]
        pdf = np.sum(pdf, axis=(-2, -1))

        if np.sum(pdf) == 0:
            return (None, None, None, None)

        # Get the reco energy bin edges and widths.
        lower_bin_edges = self.reco_e_lower_edges[
            log_true_e_idx, dec_idx
        ]
        upper_bin_edges = self.reco_e_upper_edges[
            log_true_e_idx, dec_idx
        ]
        bin_widths = upper_bin_edges - lower_bin_edges

        # Normalize the PDF.
        pdf /= np.sum(pdf) * bin_widths

        return (pdf, lower_bin_edges, upper_bin_edges, bin_widths)

    def get_psi_pdf(
            self,
            log_true_e_idx,
            dec_idx,
            log_e_idx):
        """Retrieves the psi PDF from the given true energy bin index, the
        source bin index, and the log_e bin index.
        Returns (None, None, None, None) if any of the bin indices are less then
        zero, or if the sum of all pdf bins is zero.

        Parameters
        ----------
        log_true_e_idx : int
            The index of the true energy bin.
        dec_idx : int
            The index of the declination bin.
        log_e_idx : int
            The index of the log_e bin.

        Returns
        -------
        pdf : 1d ndarray
            The psi pdf values.
        lower_bin_edges : 1d ndarray
            The lower bin edges of the psi pdf histogram.
        upper_bin_edges : 1d ndarray
            The upper bin edges of the psi pdf histogram.
        bin_widths : 1d ndarray
            The bin widths of the psi pdf histogram.
        """
        if log_true_e_idx < 0 or dec_idx < 0 or log_e_idx < 0:
            return (None, None, None, None)

        pdf = self.histogram[log_true_e_idx, dec_idx, log_e_idx]
        pdf = np.sum(pdf, axis=-1)

        if np.sum(pdf) == 0:
            return (None, None, None, None)

        # Get the PSI bin edges and widths.
        lower_bin_edges = self.psi_lower_edges[
            log_true_e_idx, dec_idx, log_e_idx
        ]
        upper_bin_edges = self.psi_upper_edges[
            log_true_e_idx, dec_idx, log_e_idx
        ]
        bin_widths = upper_bin_edges - lower_bin_edges

        # Normalize the PDF.
        pdf /= np.sum(pdf) * bin_widths

        return (pdf, lower_bin_edges, upper_bin_edges, bin_widths)

    def get_ang_err_pdf(
            self,
            log_true_e_idx,
            dec_idx,
            log_e_idx,
            psi_idx):
        """Retrieves the angular error PDF from the given true energy bin index,
        the source bin index, the log_e bin index, and the psi bin index.
        Returns (None, None, None, None) if any of the bin indices are less then
        zero, or if the sum of all pdf bins is zero.

        Parameters
        ----------
        log_true_e_idx : int
            The index of the true energy bin.
        dec_idx : int
            The index of the declination bin.
        log_e_idx : int
            The index of the log_e bin.
        psi_idx : int
            The index of the psi bin.

        Returns
        -------
        pdf : 1d ndarray
            The ang_err pdf values.
        lower_bin_edges : 1d ndarray
            The lower bin edges of the ang_err pdf histogram.
        upper_bin_edges : 1d ndarray
            The upper bin edges of the ang_err pdf histogram.
        bin_widths : 1d ndarray
            The bin widths of the ang_err pdf histogram.
        """
        if log_true_e_idx < 0 or dec_idx < 0 or log_e_idx < 0 or psi_idx < 0:
            return (None, None, None, None)

        pdf = self.histogram[log_true_e_idx, dec_idx, log_e_idx, psi_idx]

        if np.sum(pdf) == 0:
            return (None, None, None, None)

        # Get the ang_err bin edges and widths.
        lower_bin_edges = self.ang_err_lower_edges[
            log_true_e_idx, dec_idx, log_e_idx, psi_idx
        ]
        upper_bin_edges = self.ang_err_upper_edges[
            log_true_e_idx, dec_idx, log_e_idx, psi_idx
        ]
        bin_widths = upper_bin_edges - lower_bin_edges

        # Some bins might not be defined, i.e. have zero bin widths.
        valid = bin_widths > 0

        pdf = pdf[valid]
        lower_bin_edges = lower_bin_edges[valid]
        upper_bin_edges = upper_bin_edges[valid]
        bin_widths = bin_widths[valid]

        # Normalize the PDF.
        pdf = pdf / (np.sum(pdf) * bin_widths)

        return (pdf, lower_bin_edges, upper_bin_edges, bin_widths)

    def sample_log_e(
            self,
            rss,
            dec_idx,
            log_true_e_idxs):
        """Samples log energy values for the given source declination and true
        energy bins.

        Parameters
        ----------
        rss : instance of RandomStateService
            The RandomStateService which should be used for drawing random
            numbers from.
        dec_idx : int
            The index of the source declination bin.
        log_true_e_idxs : 1d ndarray of int
            The bin indices of the true energy bins.

        Returns
        -------
        log_e_idx : 1d ndarray of int
            The bin indices of the log_e pdf corresponding to the sampled
            log_e values.
        log_e : 1d ndarray of float
            The sampled log_e values.
        """
        n_evt = len(log_true_e_idxs)
        log_e_idx = np.empty((n_evt,), dtype=np.int64)
        log_e = np.empty((n_evt,), dtype=np.double)

        unique_log_true_e_idxs = np.unique(log_true_e_idxs)
        for b_log_true_e_idx in unique_log_true_e_idxs:
            m = log_true_e_idxs == b_log_true_e_idx
            b_size = np.count_nonzero(m)
            (
                pdf,
                low_bin_edges,
                up_bin_edges,
                bin_widths
            ) = self.get_log_e_pdf(
                b_log_true_e_idx,
                dec_idx)

            if pdf is None:
                log_e_idx[m] = -1
                log_e[m] = np.nan
                continue

            b_log_e_idx = rss.random.choice(
                np.arange(len(pdf)),
                p=(pdf * bin_widths),
                size=b_size)
            b_log_e = rss.random.uniform(
                low_bin_edges[b_log_e_idx],
                up_bin_edges[b_log_e_idx],
                size=b_size)

            log_e_idx[m] = b_log_e_idx
            log_e[m] = b_log_e

        return (log_e_idx, log_e)

    def sample_psi(
            self,
            rss,
            dec_idx,
            log_true_e_idxs,
            log_e_idxs):
        """Samples psi values for the given source declination, true
        energy bins, and log_e bins.

        Parameters
        ----------
        rss : instance of RandomStateService
            The RandomStateService which should be used for drawing random
            numbers from.
        dec_idx : int
            The index of the source declination bin.
        log_true_e_idxs : 1d ndarray of int
            The bin indices of the true energy bins.
        log_e_idxs : 1d ndarray of int
            The bin indices of the log_e bins.

        Returns
        -------
        psi_idx : 1d ndarray of int
            The bin indices of the psi pdf corresponding to the sampled psi
            values.
        psi : 1d ndarray of float
            The sampled psi values in radians.
        """
        if len(log_true_e_idxs) != len(log_e_idxs):
            raise ValueError(
                'The lengths of log_true_e_idxs and log_e_idxs must be equal!')

        n_evt = len(log_true_e_idxs)
        psi_idx = np.empty((n_evt,), dtype=np.int64)
        psi = np.empty((n_evt,), dtype=np.double)

        unique_log_true_e_idxs = np.unique(log_true_e_idxs)
        for b_log_true_e_idx in unique_log_true_e_idxs:
            m = log_true_e_idxs == b_log_true_e_idx
            bb_unique_log_e_idxs = np.unique(log_e_idxs[m])
            for bb_log_e_idx in bb_unique_log_e_idxs:
                mm = m & (log_e_idxs == bb_log_e_idx)
                bb_size = np.count_nonzero(mm)
                (
                    pdf,
                    low_bin_edges,
                    up_bin_edges,
                    bin_widths
                ) = self.get_psi_pdf(
                    b_log_true_e_idx,
                    dec_idx,
                    bb_log_e_idx)

                if pdf is None:
                    psi_idx[mm] = -1
                    psi[mm] = np.nan
                    continue

                bb_psi_idx = rss.random.choice(
                    np.arange(len(pdf)),
                    p=(pdf * bin_widths),
                    size=bb_size)
                bb_psi = rss.random.uniform(
                    low_bin_edges[bb_psi_idx],
                    up_bin_edges[bb_psi_idx],
                    size=bb_size)

                psi_idx[mm] = bb_psi_idx
                psi[mm] = bb_psi

        return (psi_idx, psi)

    def sample_ang_err(
            self,
            rss,
            dec_idx,
            log_true_e_idxs,
            log_e_idxs,
            psi_idxs):
        """Samples ang_err values for the given source declination, true
        energy bins, log_e bins, and psi bins.

        Parameters
        ----------
        rss : instance of RandomStateService
            The RandomStateService which should be used for drawing random
            numbers from.
        dec_idx : int
            The index of the source declination bin.
        log_true_e_idxs : 1d ndarray of int
            The bin indices of the true energy bins.
        log_e_idxs : 1d ndarray of int
            The bin indices of the log_e bins.
        psi_idxs : 1d ndarray of int
            The bin indices of the psi bins.

        Returns
        -------
        ang_err_idx : 1d ndarray of int
            The bin indices of the angular error pdf corresponding to the
            sampled angular error values.
        ang_err : 1d ndarray of float
            The sampled angular error values in radians.
        """
        if (len(log_true_e_idxs) != len(log_e_idxs)) and\
           (len(log_e_idxs) != len(psi_idxs)):
            raise ValueError(
                'The lengths of log_true_e_idxs, log_e_idxs, and psi_idxs must '
                'be equal!')

        n_evt = len(log_true_e_idxs)
        ang_err_idx = np.empty((n_evt,), dtype=np.int64)
        ang_err = np.empty((n_evt,), dtype=np.double)

        unique_log_true_e_idxs = np.unique(log_true_e_idxs)
        for b_log_true_e_idx in unique_log_true_e_idxs:
            m = log_true_e_idxs == b_log_true_e_idx
            bb_unique_log_e_idxs = np.unique(log_e_idxs[m])
            for bb_log_e_idx in bb_unique_log_e_idxs:
                mm = m & (log_e_idxs == bb_log_e_idx)
                bbb_unique_psi_idxs = np.unique(psi_idxs[mm])
                for bbb_psi_idx in bbb_unique_psi_idxs:
                    mmm = mm & (psi_idxs == bbb_psi_idx)
                    bbb_size = np.count_nonzero(mmm)
                    (
                        pdf,
                        low_bin_edges,
                        up_bin_edges,
                        bin_widths
                    ) = self.get_ang_err_pdf(
                        b_log_true_e_idx,
                        dec_idx,
                        bb_log_e_idx,
                        bbb_psi_idx)

                    if pdf is None:
                        ang_err_idx[mmm] = -1
                        ang_err[mmm] = np.nan
                        continue

                    bbb_ang_err_idx = rss.random.choice(
                        np.arange(len(pdf)),
                        p=(pdf * bin_widths),
                        size=bbb_size)
                    bbb_ang_err = rss.random.uniform(
                        low_bin_edges[bbb_ang_err_idx],
                        up_bin_edges[bbb_ang_err_idx],
                        size=bbb_size)

                    ang_err_idx[mmm] = bbb_ang_err_idx
                    ang_err[mmm] = bbb_ang_err

        return (ang_err_idx, ang_err)



# -*- coding: utf-8 -*-

"""Setup the time-dependent analysis. For now this only works on a single
dataset.
"""

import numpy as np

from skyllh.analyses.i3.publicdata_ps.backgroundpdf import (
    PDDataBackgroundI3EnergyPDF,
)
from skyllh.analyses.i3.publicdata_ps.detsigyield import (
    PDSingleParamFluxPointLikeSourceI3DetSigYieldBuilder,
)
from skyllh.analyses.i3.publicdata_ps.pdfratio import (
    PDSigSetOverBkgPDFRatio,
)
from skyllh.analyses.i3.publicdata_ps.signal_generator import (
    TimeDependentPDDatasetSignalGenerator,
)
from skyllh.analyses.i3.publicdata_ps.signalpdf import (
    PDSignalEnergyPDFSet,
)
from skyllh.analyses.i3.publicdata_ps.utils import (
    clip_grl_start_times,
    create_energy_cut_spline,
)

from skyllh.core.analysis import (
    SingleSourceMultiDatasetLLHRatioAnalysis as Analysis,
)
from skyllh.core.background_generator import (
    DatasetBackgroundGenerator,
)
from skyllh.core.backgroundpdf import (
    BackgroundTimePDF,
)
from skyllh.core.config import (
    Config,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.event_selection import (
    SpatialBoxEventSelectionMethod,
)
from skyllh.core.expectation_maximization import (
    em_fit,
)
from skyllh.core.flux_model import (
    BoxTimeFluxProfile,
    GaussianTimeFluxProfile,
    PowerLawEnergyFluxProfile,
    SteadyPointlikeFFM,
)
from skyllh.core.minimizer import (
    LBFGSMinimizerImpl,
    Minimizer,
)
from skyllh.core.minimizers.iminuit import (
    IMinuitMinimizerImpl,
)
from skyllh.core.model import (
    DetectorModel,
)
from skyllh.core.multiproc import (
    get_ncpu,
    parallelize,
)
from skyllh.core.parameters import (
    Parameter,
    ParameterModelMapper,
)
from skyllh.core.pdfratio import (
    SigOverBkgPDFRatio,
)
from skyllh.core.progressbar import (
    ProgressBar,
)
from skyllh.core.random import (
    RandomStateService,
)
from skyllh.core.scrambling import (
    DataScrambler,
)
from skyllh.core.signal_generator import (
    MultiDatasetSignalGenerator,
)
from skyllh.core.signalpdf import (
    RayleighPSFPointSourceSignalSpatialPDF,
    SignalTimePDF,
)
from skyllh.core.smoothing import (
    BlockSmoothingFilter,
)
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroup,
    SourceHypoGroupManager,
)
from skyllh.core.source_model import (
    PointLikeSource,
)
from skyllh.core.test_statistic import (
    WilksTestStatistic,
)
from skyllh.core.timing import (
    TimeLord,
)
from skyllh.core.trialdata import (
    TrialDataManager,
)
from skyllh.core.utils.analysis import (
    create_trial_data_file,
    pointlikesource_to_data_field_array,
)
from skyllh.core.utils.tdm import (
    get_tdm_field_func_psi,
)

from skyllh.datasets.i3 import (
    data_samples,
)

from skyllh.i3.background_generation import (
    FixedScrambledExpDataI3BkgGenMethod,
)
from skyllh.i3.backgroundpdf import (
    DataBackgroundI3SpatialPDF,
)
from skyllh.i3.config import (
    add_icecube_specific_analysis_required_data_fields,
)
from skyllh.i3.livetime import (
    I3Livetime,
)
from skyllh.i3.scrambling import (
    I3SeasonalVariationTimeScramblingMethod,
)

from skyllh.scripting.argparser import (
    create_argparser,
)
from skyllh.scripting.logging import (
    setup_logging,
)

TXS_0506_PLUS056_SOURCE = PointLikeSource(
    name='TXS 0506+056',
    ra=np.deg2rad(77.3581851),
    dec=np.deg2rad(5.69314828))
TXS_0506_PLUS056_ALERT_TIME = 58018.8711856


def create_signal_time_pdf(
        cfg,
        grl,
        gauss=None,
        box=None,
):
    """Creates the signal time PDF, either a gaussian or a box shaped PDF.

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    grl : instance of numpy structured ndarray
        The structured numpy ndarray holding the good-run-list data.
    gauss : dict | None
        None or dictionary with {"mu": float, "sigma": float}.
    box : dict | None
        None or dictionary with {"start": float, "stop": float}.

    Returns
    -------
    pdf : instance of PDF
        The created time PDF instance.
    """
    if (gauss is None) and (box is None):
        raise TypeError(
            'Either gauss or box have to be specified as time pdf.')

    livetime = I3Livetime.from_grl_data(
        grl_data=grl)

    if gauss is not None:
        time_flux_profile = GaussianTimeFluxProfile(
            t0=gauss['mu'],
            sigma_t=gauss['sigma'],
            cfg=cfg)
    elif box is not None:
        time_flux_profile = BoxTimeFluxProfile.from_start_and_stop_time(
            start=box['start'],
            stop=box['stop'],
            cfg=cfg)

    pdf = SignalTimePDF(
        cfg=cfg,
        livetime=livetime,
        time_flux_profile=time_flux_profile,
    )

    return pdf


def change_signal_time_pdf_of_llhratio_function(
        ana,
        gauss=None,
        box=None,
):
    """Changes the signal time PDF of the log-likelihood ratio function.

    Parameters
    ----------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The analysis instance.
    gauss : dict | None
        None or dictionary with {"mu": float, "sigma": float}.
    box : dict | None
        None or dictionary with {"start": float, "stop": float}.
    """
    cfg = ana.cfg
    grl = ana.data_list[0].grl

    time_sigpdf = create_signal_time_pdf(
        cfg=cfg,
        grl=grl,
        gauss=gauss,
        box=box)

    pdfratio = ana.llhratio.llhratio_list[0].pdfratio

    # pdfratio is an instance of PDFRatioProduct.
    # The first item is the PDF ratio product of the spatial and energy PDF
    # ratios. The second item is the time PDF ratio.
    pdfratio.pdfratio2.sig_pdf = time_sigpdf

    # TODO: Change detector signal yield with flare livetime in sample
    # (1 / grl_norm in pdf), rebuild the histograms if it is changed.


def get_energy_spatial_signal_over_background(
        ana,
        fitparam_values,
        tl=None,
):
    """Returns the signal over background ratio for
    (spatial_signal * energy_signal) / (spatial_background * energy_background).

    Parameters
    ----------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The analysis instance.
    fitparam_values : instance of ndarray
        The (N_fitparams,)-shaped numpy ndarray holding the values of the global
        fit parameters, e.g. ns and gamma.
    tl : instance of TimeLord | None
        The optional instance of TimeLord for measuring timing behavior.

    Returns
    -------
    ratio : 1d ndarray
        Product of spatial and energy signal over background pdfs.
    """
    tdm = ana.tdm_list[0]

    pdfratio = ana.llhratio.llhratio_list[0].pdfratio

    # pdfratio is an instance of PDFRatioProduct.
    # The first item is the PDF ratio product of the spatial and energy PDF
    # ratios. The second item is the time PDF ratio.
    pdfratio = pdfratio.pdfratio1

    src_params_recarray = ana.pmm.create_src_params_recarray(
        gflp_values=fitparam_values)

    ratio = pdfratio.get_ratio(
        tdm=tdm,
        src_params_recarray=src_params_recarray,
        tl=tl)

    return ratio


def change_fluxmodel_gamma(
        ana,
        gamma,
):
    """Sets the given gamma value to the flux model of the single source.

    Parameters
    ----------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The analysis that should be used.
    gamma : float
        Spectral index for the flux model.
    """
    ana.shg_mgr.shg_list[0].fluxmodel.set_params({'gamma': gamma})
    ana.change_shg_mgr(shg_mgr=ana.shg_mgr)


def change_time_flux_profile_params(
        ana,
        params,
):
    """Changes the parameters of the source's time flux profile.

    Parameters
    ----------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The analysis that should be used.
    params : dict
        The dictionary with the parameter names and values to be set.
    """
    # Note: In the future the primary storage place for the time flux profile
    #       will be within the flux model.
    ana.sig_generator_list[0]._time_flux_profile.set_params(params)


def calculate_TS(
        ana,
        em_results,
        rss,
):
    """Calculate the best TS value from the expectation maximization gamma scan
    results.

    Parameters
    ----------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The analysis that should be used.
    em_results : instance of structured ndarray
        The numpy structured ndarray holding the EM results (from the gamma
        scan).
    rss : instance of RandomStateService
        The instance of RandomStateService that should be used to generate
        random numbers from.

    Returns
    -------
    max_TS : float
        The maximal TS value of all maximized time hypotheses.
    best_em_result : instance of numpy structured ndarray
        The row of ``em_results`` that corresponds to the best fit.
    best_fitparam_values : instance of numpy ndarray
        The instance of numpy ndarray holding the fit parameter values of the
        overall best fit result.
    """
    max_TS = None
    best_em_result = None
    best_fitparam_values = None
    for em_result in em_results:
        change_signal_time_pdf_of_llhratio_function(
            ana=ana,
            gauss={
                'mu': em_result['mu'],
                'sigma': em_result['sigma']})

        (log_lambda_max, fitparam_values, status) = ana.llhratio.maximize(
            rss=rss)

        TS = ana.calculate_test_statistic(
            log_lambda=log_lambda_max,
            fitparam_values=fitparam_values)

        if (max_TS is None) or (TS > max_TS):
            max_TS = TS
            best_em_result = em_result
            best_fitparam_values = fitparam_values

    return (max_TS, best_em_result, best_fitparam_values)


def run_gamma_scan_for_single_flare(
        ana,
        remove_time=None,
        gamma_min=1,
        gamma_max=5,
        n_gamma=51,
        ppbar=None,
):
    """Runs ``em_fit`` for different gamma values in the signal energy PDF.

    Parameters
    ----------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The analysis that should be used.
    remove_time : float
        Time information of event that should be removed.
    gamma_min : float
        Lower bound for gamma scan.
    gamma_max : float
        Upper bound for gamma scan.
    n_gamma : int
        Number of steps for gamma scan.
    ppbar : instance of ProgressBar | None
        The optional parent instance of ProgressBar.

    Returns
    -------
    em_results : instance of numpy structured ndarray
        The numpy structured ndarray with fields

        gamma : float
            The spectral index value.
        mu : float
            The determined mean value of the gauss curve.
        sigma : float
            The determined standard deviation of the gauss curve.
        ns_em : float
            The scaling factor of the flare.
    """
    em_results_dt = [
        ('gamma', np.float64),
        ('mu', np.float64),
        ('sigma', np.float64),
        ('ns_em', np.float64),
    ]
    em_results = np.empty(n_gamma, dtype=em_results_dt)

    time = ana._tdm_list[0].get_data('time')

    gamma_values = np.linspace(gamma_min, gamma_max, n_gamma)

    pbar = ProgressBar(len(gamma_values), parent=ppbar).start()
    for (i, gamma) in enumerate(gamma_values):
        fitparam_values = np.array([0, gamma], dtype=np.float64)
        ratio = get_energy_spatial_signal_over_background(ana, fitparam_values)
        (mu, sigma, ns) = em_fit(
            time,
            ratio,
            n=1,
            tol=1.e-200,
            iter_max=500,
            weight_thresh=0,
            initial_width=5000,
            remove_x=remove_time)
        em_results[i] = (
            gamma,
            mu[0],
            sigma[0],
            ns[0],
        )
        pbar.increment()
    pbar.finish()

    return em_results


def unblind_single_flare(
        ana,
        remove_time=None,
):
    """Run EM for a single flare on unblinded data.
    Similar to the original analysis, remove the alert event.

    Parameters
    ----------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The analysis that should be used.
    remove_time : float
        Time of the event that should be removed.
        In the case of the TXS analysis:
        ``remove_time=TXS_0506_PLUS056_ALERT_TIME``.

    Returns
    -------
    max_TS : float
        The maximal TS value of all maximized time hypotheses.
    best_em_result : instance of numpy structured ndarray
        The EM result from the gamma scan corresponding to the best fit.
    best_fitparam_values : instance of numpy ndarray
        The instance of numpy ndarray holding the fit parameter values of the
        overall best fit result.
    """
    rss = RandomStateService(seed=1)

    ana.unblind(
        minimizer_rss=rss)

    em_results = run_gamma_scan_for_single_flare(
        ana=ana,
        remove_time=remove_time)

    (max_ts, best_em_result, best_fitparam_values) = calculate_TS(
        ana=ana,
        em_results=em_results,
        rss=rss)

    return (max_ts, best_em_result, best_fitparam_values)


def do_trial_with_em(
        ana,
        rss,
        mean_n_sig=0,
        gamma_src=2,
        gamma_min=1,
        gamma_max=5,
        n_gamma=21,
        gauss=None,
        box=None,
        tl=None,
        ppbar=None,
):
    """Performs a trial using the expectation maximization algorithm.
    It runs a gamma scan and does the EM for each gamma value.

    Parameters
    ----------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The analysis instance that should be used to perform the trial.
    rss : instance of RandomStateService
        The instance of RandomStateService that should be used to generate
        random numbers.
    mean_n_sig : float
        The mean number of signal events that should be generated.
    gamma_src : float
        The spectral index of the source.
    gamma_min : float
        Lower bound of the gamma scan.
    gamma_max : float
        Upper bound of the gamma scan.
    n_gamma : int
        Number of steps of the gamma scan.
    gauss : dict | None
        Properties of the Gaussian time PDF.
        None or dictionary with {"mu": float, "sigma": float}.
    box : dict | None
        Properties of the box time PDF.
        None or dictionary with {"start": float, "stop": float}.
    tl : instance of TimeLord | None
        The optional instance of TimeLord to measure timing information.
    ppbar : instance of ProgressBar | None
        The optional parent instance of ProgressBar.

    Returns
    -------
    trial : instance of structured ndarray
        The numpy structured ndarray of length 1 with the following fields:

        seed : numpy.int64
            The seed value used to generate the trial.
        mean_n_sig : numpy.float64
            The mean number of signal events of the trial.
        n_sig : numpy.int64
            The actual number of signal events in the trial.
        gamma_src : numpy.float64
            The spectral index of the source.
        mu_sig : numpy.float64
            The mean value of the Gaussian time PDF of the source.
        sigma_sig : numpy.float64
            The sigma value of the Gaussian time PDF of the source.
        start_sig : numpy.float64
            The start time of the box time PDF of the source.
        stop_sig : numpy.float64
            The stop time of the box time PDF of the source.
        ts : numpy.float64
            The test-statistic value of the trial.
        ns_fit : numpy.float64
            The fitted number of signal events.
        ns_em : numpy.float64
            The scaling factor of the flare.
        gamma_fit : numpy.float64
            The fitted spectral index of the trial.
        gamma_em : numpy.float64
            The spectral index of the best EM trial.
        mu_fit : numpy.float64
            The fitted mean value of the Gaussian time PDF.
        sigma_fit : numpy.float64
            The fitted sigma value of the Gaussian time PDF.
    """
    trial_dt = [
        ('seed', np.int64),
        ('mean_n_sig', np.float64),
        ('n_sig', np.int64),
        ('gamma_src', np.float64),
        ('mu_sig', np.float64),
        ('sigma_sig', np.float64),
        ('start_sig', np.float64),
        ('stop_sig', np.float64),
        ('ts', np.float64),
        ('ns_fit', np.float64),
        ('ns_em', np.float64),
        ('gamma_fit', np.float64),
        ('gamma_em', np.float64),
        ('mu_fit', np.float64),
        ('sigma_fit', np.float64)
    ]

    trial = np.empty((1,), dtype=trial_dt)

    (n_sig, n_events_list, events_list) = ana.generate_pseudo_data(
        rss=rss,
        mean_n_sig=mean_n_sig,
        tl=tl)
    ana.initialize_trial(events_list, n_events_list)

    em_results = run_gamma_scan_for_single_flare(
        ana=ana,
        gamma_min=gamma_min,
        gamma_max=gamma_max,
        n_gamma=n_gamma,
        ppbar=ppbar)

    (max_ts, best_em_result, best_fitparams) = calculate_TS(
        ana=ana,
        em_results=em_results,
        rss=rss)

    trial[0] = (
        rss.seed,
        mean_n_sig,
        n_sig,
        gamma_src,
        gauss['mu'] if gauss is not None else -1,
        gauss['sigma'] if gauss is not None else -1,
        box['start'] if box is not None else -1,
        box['end'] if box is not None else -1,
        max_ts,
        best_fitparams[0],
        best_em_result['ns_em'],
        best_fitparams[1],
        best_em_result['gamma'],
        best_em_result['mu'],
        best_em_result['sigma']
    )

    return trial


def do_trials_with_em(
        ana,
        n=1000,
        ncpu=None,
        seed=1,
        mean_n_sig=0,
        gamma_src=2,
        gamma_min=1,
        gamma_max=5,
        n_gamma=21,
        gauss=None,
        box=None,
        tl=None,
        ppbar=None,
):
    """Performs ``n_trials`` trials using the expectation maximization
    algorithm. For each trial it runs a gamma scan and does the EM for each
    gamma value.

    Parameters
    ----------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The analysis instance that should be used to perform the trials.
    n : int
        The number of trials to generate.
    ncpu : int | None
        The number of CPUs to use to generate the trials. If set to ``None``
        the configured default value will be used.
    mean_n_sig : float
        The mean number of signal events that should be generated.
    gamma_src : float
        The spectral index of the source.
    gamma_min : float
        Lower bound of the gamma scan.
    gamma_max : float
        Upper bound of the gamma scan.
    n_gamma : int
        Number of steps of the gamma scan.
    seed : int
        The seed for the random number generator.
    gauss : dict | None
        Properties of the Gaussian time PDF.
        None or dictionary with {"mu": float, "sigma": float}.
    box : dict | None
        Properties of the box time PDF.
        None or dictionary with {"start": float, "stop": float}.
    tl : instance of TimeLord | None
        The optional instance of TimeLord to measure timing information.
    ppbar : instance of ProgressBar | None
        The optional parent instance of ProgressBar.

    Returns
    -------
    trials : instance of numpy structured ndarray
        The numpy structured ndarray of length ``n_trials`` with the results for
        each trial. The array has the following fields:

        seed : numpy.int64
            The seed value used to generate the trial.
        mean_n_sig : numpy.float64
            The mean number of signal events of the trial.
        n_sig : numpy.int64
            The actual number of signal events in the trial.
        gamma_src : numpy.float64
            The spectral index of the source.
        mu_sig : numpy.float64
            The mean value of the Gaussian time PDF of the source.
        sigma_sig : numpy.float64
            The sigma value of the Gaussian time PDF of the source.
        start_sig : numpy.float64
            The start time of the box time PDF of the source.
        stop_sig : numpy.float64
            The stop time of the box time PDF of the source.
        ts : numpy.float64
            The test-statistic value of the trial.
        ns_fit : numpy.float64
            The fitted number of signal events.
        ns_em : numpy.float64
            The scaling factor of the flare.
        gamma_fit : numpy.float64
            The fitted spectral index of the trial.
        gamma_em : numpy.float64
            The spectral index of the best EM trial.
        mu_fit : numpy.float64
            The fitted mean value of the Gaussian time PDF.
        sigma_fit : numpy.float64
            The fitted sigma value of the Gaussian time PDF.
    """
    rss = RandomStateService(seed=seed)

    if mean_n_sig > 0:
        change_fluxmodel_gamma(
            ana=ana,
            gamma=gamma_src)

    args_list = [
        (
            tuple(),
            dict(
                ana=ana,
                mean_n_sig=mean_n_sig,
                gamma_src=gamma_src,
                gamma_min=gamma_min,
                gamma_max=gamma_max,
                n_gamma=n_gamma,
                gauss=gauss,
                box=box,
                ppbar=False,
            )
        )
        for i in range(n)
    ]

    result_list = parallelize(
        func=do_trial_with_em,
        args_list=args_list,
        ncpu=get_ncpu(cfg=ana.cfg, local_ncpu=ncpu),
        rss=rss,
        tl=tl,
        ppbar=ppbar,
    )

    trials = None
    for (i, result) in enumerate(result_list):
        if trials is None:
            trials = np.empty((n,), dtype=result.dtype)
        trials[i] = result[0]

    return trials


def create_analysis(  # noqa: C901
        cfg,
        datasets,
        source,
        box=None,
        gauss=None,
        refplflux_Phi0=1,
        refplflux_E0=1e3,
        refplflux_gamma=2.0,
        ns_seed=10.0,
        ns_min=0.,
        ns_max=1e3,
        gamma_seed=3.0,
        gamma_min=1.,
        gamma_max=5.,
        kde_smoothing=False,
        minimizer_impl="LBFGS",
        cut_sindec=None,
        spl_smooth=None,
        cap_ratio=False,
        compress_data=False,
        keep_data_fields=None,
        evt_sel_delta_angle_deg=10,
        construct_bkg_generator=True,
        construct_sig_generator=True,
        tl=None,
        ppbar=None,
        logger_name=None,
):
    """Creates the Analysis instance for this particular analysis.

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    datasets : list of Dataset instances
        The list of Dataset instances, which should be used in the
        analysis.
    source : PointLikeSource instance
        The PointLikeSource instance defining the point source position.
    box : None or dictionary with start, end
        None if no box shaped time pdf, else dictionary of the format
        ``{'start': float, 'stop': float}``.
    gauss : None or dictionary with mu, sigma
        None if no gaussian time pdf, else dictionary of the format
        ``{'mu': float, 'sigma': float}``.
    refplflux_Phi0 : float
        The flux normalization to use for the reference power law flux model.
    refplflux_E0 : float
        The reference energy to use for the reference power law flux model.
    refplflux_gamma : float
        The spectral index to use for the reference power law flux model.
    ns_seed : float
        Value to seed the minimizer with for the ns fit.
    ns_min : float
        Lower bound for ns fit.
    ns_max : float
        Upper bound for ns fit.
    gamma_seed : float | None
        Value to seed the minimizer with for the gamma fit. If set to None,
        the refplflux_gamma value will be set as gamma_seed.
    gamma_min : float
        Lower bound for gamma fit.
    gamma_max : float
        Upper bound for gamma fit.
    kde_smoothing : bool
        Apply a KDE-based smoothing to the data-driven background pdf.
        Default: False.
    minimizer_impl : str | "LBFGS"
        Minimizer implementation to be used. Supported options are "LBFGS"
        (L-BFG-S minimizer used from the :mod:`scipy.optimize` module), or
        "minuit" (Minuit minimizer used by the :mod:`iminuit` module).
        Default: "LBFGS".
    cut_sindec : list of float | None
        sin(dec) values at which the energy cut in the southern sky should
        start. If None, np.sin(np.radians([-2, 0, -3, 0, 0])) is used.
    spl_smooth : list of float
        Smoothing parameters for the 1D spline for the energy cut. If None,
        [0., 0.005, 0.05, 0.2, 0.3] is used.
    cap_ratio : bool
        If set to True, the energy PDF ratio will be capped to a finite value
        where no background energy PDF information is available. This will
        ensure that an energy PDF ratio is available for high energies where
        no background is available from the experimental data.
        If kde_smoothing is set to True, cap_ratio should be set to False!
        Default is False.
    compress_data : bool
        Flag if the data should get converted from float64 into float32.
    keep_data_fields : list of str | None
        List of additional data field names that should get kept when loading
        the data.
    evt_sel_delta_angle_deg : float
        The delta angle in degrees for the event selection optimization methods.
    construct_bkg_generator : bool
        Flag if the background generator should be constructed (``True``) or not
        (``False``).
    construct_sig_generator : bool
        Flag if the signal generator should be constructed (``True``) or not
        (``False``).
    tl : TimeLord instance | None
        The TimeLord instance to use to time the creation of the analysis.
    ppbar : ProgressBar instance | None
        The instance of ProgressBar for the optional parent progress bar.
    logger_name : str | None
        The name of the logger to be used. If set to ``None``, ``__name__`` will
        be used.

    Returns
    -------
    ana : instance of SingleSourceMultiDatasetLLHRatioAnalysis
        The Analysis instance for this analysis.
    """
    add_icecube_specific_analysis_required_data_fields(cfg)

    # Remove run number from the dataset data field requirements.
    cfg['datafields'].pop('run', None)

    if logger_name is None:
        logger_name = __name__
    logger = get_logger(logger_name)

    if len(datasets) != 1:
        raise RuntimeError(
            'This analysis supports only analyses with only single datasets '
            'at the moment!')

    if (gauss is None) and (box is None):
        raise ValueError(
            'No time pdf specified (box or gauss)!')
    if (gauss is not None) and (box is not None):
        raise ValueError(
            'Time PDF cannot be both gaussian and box shaped. '
            'Please specify only one shape.')

    # Create the minimizer instance.
    if minimizer_impl == 'LBFGS':
        minimizer = Minimizer(LBFGSMinimizerImpl(cfg=cfg))
    elif minimizer_impl == 'minuit':
        minimizer = Minimizer(IMinuitMinimizerImpl(cfg=cfg, ftol=1e-8))
    else:
        raise NameError(
            f"Minimizer implementation `{minimizer_impl}` is not supported "
            "Please use `LBFGS` or `minuit`.")

    dtc_dict = None
    dtc_except_fields = None
    if compress_data is True:
        dtc_dict = {np.dtype(np.float64): np.dtype(np.float32)}
        dtc_except_fields = ['mcweight']

    # Define the flux model.
    fluxmodel = SteadyPointlikeFFM(
        Phi0=refplflux_Phi0,
        energy_profile=PowerLawEnergyFluxProfile(
            E0=refplflux_E0,
            gamma=refplflux_gamma,
            cfg=cfg,
        ),
        cfg=cfg,
    )

    # Define the time flux profile of the source.
    time_flux_profile = None
    if box is not None:
        time_flux_profile = BoxTimeFluxProfile.from_start_and_stop_time(
            start=box['start'],
            stop=box['stop'],
            cfg=cfg)
    elif gauss is not None:
        time_flux_profile = GaussianTimeFluxProfile(
            t0=gauss['mu'],
            sigma_t=gauss['sigma'],
            cfg=cfg)

    # Define the fit parameter ns.
    param_ns = Parameter(
        name='ns',
        initial=ns_seed,
        valmin=ns_min,
        valmax=ns_max)

    # Define the fit parameter gamma.
    param_gamma = Parameter(
        name='gamma',
        initial=gamma_seed,
        valmin=gamma_min,
        valmax=gamma_max)

    # Define the detector signal yield builder for the IceCube detector and this
    # source and flux model.
    # The sin(dec) binning will be taken by the builder automatically from the
    # Dataset instance.
    gamma_grid = param_gamma.as_linear_grid(delta=0.1)
    detsigyield_builder =\
        PDSingleParamFluxPointLikeSourceI3DetSigYieldBuilder(
            cfg=cfg,
            param_grid=gamma_grid)

    # Create a source hypothesis group manager with a single source hypothesis
    # group for the single source.
    shg_mgr = SourceHypoGroupManager(
        SourceHypoGroup(
            sources=source,
            fluxmodel=fluxmodel,
            detsigyield_builders=detsigyield_builder,
        ))
    logger.info(str(shg_mgr))

    # Define a detector model for the ns fit parameter.
    detector_model = DetectorModel('IceCube')

    # Define the parameter model mapper for the analysis, which will map global
    # parameters to local source parameters.
    pmm = ParameterModelMapper(
        models=[detector_model, source])
    pmm.map_param(param_ns, models=detector_model)
    pmm.map_param(param_gamma, models=source)
    logger.info(str(pmm))

    # Define the test statistic.
    test_statistic = WilksTestStatistic()

    # Create the Analysis instance.
    ana = Analysis(
        cfg=cfg,
        shg_mgr=shg_mgr,
        pmm=pmm,
        test_statistic=test_statistic,
        sig_generator_cls=MultiDatasetSignalGenerator,
    )

    # Define the event selection method for pure optimization purposes.
    # We will use the same method for all datasets.
    event_selection_method = SpatialBoxEventSelectionMethod(
        shg_mgr=shg_mgr,
        delta_angle=np.deg2rad(evt_sel_delta_angle_deg))

    # Prepare the spline parameters for the signal generator.
    if cut_sindec is None:
        cut_sindec = np.sin(np.radians([-2, 0, -3, 0, 0]))
    if spl_smooth is None:
        spl_smooth = [0., 0.005, 0.05, 0.2, 0.3]
    if len(spl_smooth) < len(datasets) or len(cut_sindec) < len(datasets):
        raise AssertionError(
            'The length of the spl_smooth and of the cut_sindec must be equal '
            f'to the length of datasets: {len(datasets)}.')

    # Add the data sets to the analysis.
    pbar = ProgressBar(len(datasets), parent=ppbar).start()
    for (ds_idx, ds) in enumerate(datasets):
        data = ds.load_and_prepare_data(
            keep_fields=keep_data_fields,
            dtc_dict=dtc_dict,
            dtc_except_fields=dtc_except_fields,
            tl=tl)

        # Some runs might overlap slightly. So we need to clip those runs.
        clip_grl_start_times(grl_data=data.grl)

        livetime = I3Livetime.from_grl_data(
            grl_data=data.grl)

        sin_dec_binning = ds.get_binning_definition('sin_dec')
        log_energy_binning = ds.get_binning_definition('log_energy')

        # Create the spatial PDF ratio instance for this dataset.
        spatial_sigpdf = RayleighPSFPointSourceSignalSpatialPDF(
            cfg=cfg,
            dec_range=np.arcsin(sin_dec_binning.range))
        spatial_bkgpdf = DataBackgroundI3SpatialPDF(
            cfg=cfg,
            data_exp=data.exp,
            sin_dec_binning=sin_dec_binning)
        spatial_pdfratio = SigOverBkgPDFRatio(
            cfg=cfg,
            sig_pdf=spatial_sigpdf,
            bkg_pdf=spatial_bkgpdf)

        # Create the energy PDF ratio instance for this dataset.
        energy_sigpdfset = PDSignalEnergyPDFSet(
            cfg=cfg,
            ds=ds,
            src_dec=source.dec,
            fluxmodel=fluxmodel,
            param_grid_set=gamma_grid,
            ppbar=ppbar
        )
        smoothing_filter = BlockSmoothingFilter(nbins=1)
        energy_bkgpdf = PDDataBackgroundI3EnergyPDF(
            cfg=cfg,
            data_exp=data.exp,
            logE_binning=log_energy_binning,
            sinDec_binning=sin_dec_binning,
            smoothing_filter=smoothing_filter,
            kde_smoothing=kde_smoothing)

        energy_pdfratio = PDSigSetOverBkgPDFRatio(
            cfg=cfg,
            sig_pdf_set=energy_sigpdfset,
            bkg_pdf=energy_bkgpdf,
            cap_ratio=cap_ratio)

        pdfratio = spatial_pdfratio * energy_pdfratio

        # Create the time PDF ratio instance for this dataset.
        if (gauss is not None) or (box is not None):
            time_bkgpdf = BackgroundTimePDF(
                cfg=cfg,
                livetime=livetime,
                time_flux_profile=BoxTimeFluxProfile.from_start_and_stop_time(
                    start=livetime.time_start,
                    stop=livetime.time_stop,
                    cfg=cfg))
            time_sigpdf = create_signal_time_pdf(
                cfg=cfg,
                grl=data.grl,
                gauss=gauss,
                box=box)
            time_pdfratio = SigOverBkgPDFRatio(
                cfg=cfg,
                sig_pdf=time_sigpdf,
                bkg_pdf=time_bkgpdf,
                same_axes=False,
            )

            pdfratio = pdfratio * time_pdfratio

        # Create a trial data manager and add the required data fields.
        tdm = TrialDataManager()
        tdm.add_source_data_field(
            name='src_array',
            func=pointlikesource_to_data_field_array)
        tdm.add_data_field(
            name='psi',
            func=get_tdm_field_func_psi(),
            dt='dec',
            is_srcevt_data=True)

        data_scrambler = DataScrambler(
            I3SeasonalVariationTimeScramblingMethod(
                data=data))
        bkg_gen_method = FixedScrambledExpDataI3BkgGenMethod(
            cfg=cfg,
            data_scrambler=data_scrambler)

        bkg_generator = DatasetBackgroundGenerator(
            cfg=cfg,
            dataset=ds,
            data=data,
            bkg_gen_method=bkg_gen_method,
        )

        energy_cut_spline = create_energy_cut_spline(
            ds,
            data.exp,
            spl_smooth[ds_idx])

        sig_generator = TimeDependentPDDatasetSignalGenerator(
            cfg=cfg,
            shg_mgr=shg_mgr,
            ds=ds,
            ds_idx=ds_idx,
            livetime=livetime,
            time_flux_profile=time_flux_profile,
            energy_cut_spline=energy_cut_spline,
            cut_sindec=cut_sindec[ds_idx],
        )

        ana.add_dataset(
            ds,
            data,
            pdfratio=pdfratio,
            tdm=tdm,
            event_selection_method=event_selection_method,
            bkg_generator=bkg_generator,
            sig_generator=sig_generator)

        pbar.increment()
    pbar.finish()

    ana.construct_services(
        ppbar=ppbar)

    ana.llhratio = ana.construct_llhratio(
        minimizer=minimizer,
        ppbar=ppbar)

    if construct_bkg_generator is True:
        ana.construct_background_generator()

    if construct_sig_generator is True:
        ana.construct_signal_generator()

    return ana


if __name__ == '__main__':
    parser = create_argparser(
        description='Calculates TS for a given source location using the '
                    '10-year public point source sample assuming a signal '
                    'time PDF.',
    )

    parser.add_argument(
        '--dec',
        dest='dec',
        default=5.7,
        type=float,
        help='The source declination in degrees.'
    )
    parser.add_argument(
        '--ra',
        dest='ra',
        default=77.35,
        type=float,
        help='The source right-ascension in degrees.'
    )
    parser.add_argument(
        '--gamma-seed',
        dest='gamma_seed',
        default=3,
        type=float,
        help='The seed value of the gamma fit parameter.'
    )

    args = parser.parse_args()

    cfg = Config.from_yaml(args.config)
    cfg.set_enable_tracing(args.enable_tracing)
    cfg.set_ncpu(args.n_cpu)

    setup_logging(
        cfg=cfg,
        script_logger_name=__name__,
        debug_pathfilename=args.debug_logfile)

    sample_seasons = [
        ('PublicData_10y_ps', 'IC86_II-VII'),
    ]

    datasets = []
    for (sample, season) in sample_seasons:
        # Get the dataset from the correct dataset collection.
        dsc = data_samples[sample].create_dataset_collection(
            cfg=cfg,
            base_path=args.data_basepath)
        datasets.append(dsc.get_dataset(season))

    # Define a random state service.
    rss = RandomStateService(args.seed)

    # Define the point source.
    source = PointLikeSource(
        name='My Point-Like-Source',
        ra=np.deg2rad(args.ra),
        dec=np.deg2rad(args.dec))
    print(f'source: {source}')

    tl = TimeLord()

    with tl.task_timer('Creating analysis.'):
        ana = create_analysis(
            cfg=cfg,
            datasets=datasets,
            source=source,
            gamma_seed=args.gamma_seed,
            gauss={'mu': 57000, 'sigma': 62},
            tl=tl)

    with tl.task_timer('Unblinding data.'):
        (TS, param_dict, status) = ana.unblind(
            minimizer_rss=rss)

    print(f'TS = {TS:g}')
    print(f'ns_fit = {param_dict["ns"]:g}')
    print(f'gamma_fit = {param_dict["gamma"]:g}')
    print(f'minimizer status = {status}')

    print(tl)

    tl = TimeLord()
    rss = RandomStateService(seed=1)
    (_, _, _, trials) = create_trial_data_file(
        ana=ana,
        rss=rss,
        n_trials=10,
        mean_n_sig=0,
        pathfilename=None,
        ncpu=1,
        tl=tl)
    print(f'trials: {trials}')
    print(tl)



# -*- coding: utf-8 -*-

"""This module contains utility functions for creating and managing
MultiDimGridPDF instances.
"""

import numpy as np

from skyllh.core.binning import (
    BinningDefinition,
)
from skyllh.core.pdf import (
    MultiDimGridPDF,
)
from skyllh.core.py import (
    classname,
)


def get_kde_pdf_sig_spatial_norm_factor_func(
        log10_psi_name='log10_psi'):
    """Returns the standard normalization factor function for the spatial
    signal MultiDimGridPDF, which is created from KDE PDF values.
    It can be used for the ``norm_factor_func`` argument of the
    ``create_MultiDimGridPDF_from_photosplinetable`` and
    ``create_MultiDimGridPDF_from_kde_pdf`` function.

    Parameters
    ----------
    log10_psi_name : str
        The name of the event data field for the log10(psi) values.
    """
    def kde_pdf_sig_spatial_norm_factor_func(
            pdf,
            tdm,
            params_recarray,
            eventdata,
            evt_mask=None):

        log10_psi_idx = pdf._axes.get_index_by_name(log10_psi_name)

        if evt_mask is None:
            psi = 10**eventdata[log10_psi_idx]
        else:
            psi = 10**eventdata[log10_psi_idx][evt_mask]

        norm = 1. / (2 * np.pi * np.log(10) * psi * np.sin(psi))

        return norm

    return kde_pdf_sig_spatial_norm_factor_func


def get_kde_pdf_bkg_norm_factor_func():
    """Returns the standard normalization factor function for the background
    MultiDimGridPDF, which is created from KDE PDF values.
    It can be used for the ``norm_factor_func`` argument of the
    ``create_MultiDimGridPDF_from_photosplinetable`` and
    ``create_MultiDimGridPDF_from_kde_pdf`` function.
    """
    def kde_pdf_bkg_norm_factor_func(
            pdf,
            tdm,
            params_recarray,
            eventdata,
            evt_mask=None):

        return 1. / (2 * np.pi)

    return kde_pdf_bkg_norm_factor_func


def create_MultiDimGridPDF_from_photosplinetable(
        multidimgridpdf_cls,
        pmm,
        ds,
        data,
        info_key,
        splinetable_key,
        kde_pdf_axis_name_map_key='KDE_PDF_axis_name_map',
        norm_factor_func=None,
        cache_pd_values=False,
        tl=None,
        **kwargs,
):
    """
    Creates a MultiDimGridPDF instance with pdf values taken from a photospline
    pdf, i.e. a spline interpolation of KDE PDF values stored in a splinetable
    on disk.

    Parameters
    ----------
    multidimgridpdf_cls : subclass of MultiDimGridPDF
        The MultiDimGridPDF class, which should be used.
    pmm : instance of ParameterModelMapper
        The instance of ParameterModelMapper, which defines the mapping of
        global parameters to local model parameters.
    ds : instance of Dataset
        The instance of Dataset the PDF applies to.
    data : instance of DatasetData
        The instance of DatasetData that holds the experimental and monte-carlo
        data of the dataset.
    info_key : str
        The auxiliary data name for the file containing PDF information.
    splinetable_key : str
        The auxiliary data name for the name of the file containing the
        photospline spline table.
    kde_pdf_axis_name_map_key : str
        The auxiliary data name for the KDE PDF axis name map.
    norm_factor_func : callable | None
        The function that calculates a possible required normalization
        factor for the PDF value based on the event properties.
        For more information about this argument see the documentation of the
        :meth:`skyllh.core.pdf.MultiDimGridPDF.__init__` method.
    cache_pd_values : bool
        Flag if the probability density values should get cached by the
        MultiDimGridPDF class.
    tl : instance of TimeLord | None
        The optional instance of TimeLord to use for measuring timing
        information.

    Returns
    -------
    pdf : instance of ``multidimgridpdf_cls``
        The created PDF instance of MultiDimGridPDF.
    """
    if not issubclass(multidimgridpdf_cls, MultiDimGridPDF):
        raise TypeError(
            'The multidimgridpdf_cls argument must be a subclass of '
            'MultiDimGridPDF! '
            f'Its current type is {classname(multidimgridpdf_cls)}.')

    # Load the PDF data from the auxilary files.
    num_dict = ds.load_aux_data(info_key, tl=tl)

    kde_pdf_axis_name_map = ds.load_aux_data(kde_pdf_axis_name_map_key, tl=tl)
    kde_pdf_axis_name_map_inv = dict(
        [(v, k) for (k, v) in kde_pdf_axis_name_map.items()])
    for var in num_dict['vars']:
        if var not in kde_pdf_axis_name_map_inv:
            kde_pdf_axis_name_map_inv[var] = var

    if 'bin_centers' in num_dict:
        bin_centers_key = 'bin_centers'
    elif 'bins' in num_dict:
        bin_centers_key = 'bins'
    else:
        raise KeyError(
            'The PDF information file is missing "bin_centers" or "bins" key!')

    axis_binnings = [
        BinningDefinition(
            name=kde_pdf_axis_name_map_inv[var],
            binedges=num_dict[bin_centers_key][idx])
        for (idx, var) in enumerate(num_dict['vars'])
    ]

    # Getting the name of the splinetable file
    splinetable_file = ds.get_abs_pathfilename_list(
        ds.get_aux_data_definition(splinetable_key))[0]

    pdf = multidimgridpdf_cls(
        pmm=pmm,
        axis_binnings=axis_binnings,
        path_to_pdf_splinetable=splinetable_file,
        norm_factor_func=norm_factor_func,
        cache_pd_values=cache_pd_values,
        **kwargs)

    return pdf


def create_MultiDimGridPDF_from_kde_pdf(  # noqa: C901
        multidimgridpdf_cls,
        pmm,
        ds,
        data,
        numerator_key,
        denumerator_key=None,
        kde_pdf_axis_name_map_key='KDE_PDF_axis_name_map',
        norm_factor_func=None,
        cache_pd_values=False,
        tl=None,
        **kwargs,
):
    """Creates a MultiDimGridPDF instance with pdf values taken from KDE PDF
    values stored in the dataset's auxiliary data.

    Parameters
    ----------
    multidimgridpdf_cls : subclass of MultiDimGridPDF
        The MultiDimGridPDF class, which should be used.
    pmm : instance of ParameterModelMapper
        The instance of ParameterModelMapper, which defines the mapping of
        global parameters to local model parameters.
    ds : instance of Dataset
        The instance of Dataset the PDF applies to.
    data : instance of DatasetData
        The instance of DatasetData that holds the auxiliary data of the
        dataset.
    numerator_key : str
        The auxiliary data name for the PDF numerator array.
    denumerator_key : str | None
        The auxiliary data name for the PDF denumerator array.
        This can be ``None``, if no denumerator array is required.
    kde_pdf_axis_name_map_key : str
        The auxiliary data name for the KDE PDF axis name map.
    norm_factor_func : callable | None
        The function that calculates a possible required normalization
        factor for the PDF value based on the event properties.
        For more information about this argument see the documentation of the
        :meth:`skyllh.core.pdf.MultiDimGridPDF.__init__` method.
    cache_pd_values : bool
        Flag if the probability density values should get cached by the
        MultiDimGridPDF class.
    tl : instance of TimeLord | None
        The optional instance of TimeLord to use for measuring timing
        information.

    Returns
    -------
    pdf : instance of ``multidimgridpdf_cls``
        The created PDF instance of MultiDimGridPDF.
    """
    if not issubclass(multidimgridpdf_cls, MultiDimGridPDF):
        raise TypeError(
            'The multidimgridpdf_cls argument must be a subclass of '
            'MultiDimGridPDF! '
            f'Its current type is {classname(multidimgridpdf_cls)}.')

    # Load the PDF data from the auxilary files.
    num_dict = ds.load_aux_data(numerator_key, tl=tl)

    denum_dict = None
    if denumerator_key is not None:
        denum_dict = ds.load_aux_data(denumerator_key, tl=tl)

    kde_pdf_axis_name_map = ds.load_aux_data(kde_pdf_axis_name_map_key, tl=tl)
    kde_pdf_axis_name_map_inv = dict(
        [(v, k) for (k, v) in kde_pdf_axis_name_map.items()])
    for var in num_dict['vars']:
        if var not in kde_pdf_axis_name_map_inv:
            kde_pdf_axis_name_map_inv[var] = var

    if 'bin_centers' in num_dict:
        bin_centers_key = 'bin_centers'
    elif 'bins' in num_dict:
        bin_centers_key = 'bins'
    else:
        raise KeyError(
            'The PDF information file is missing "bin_centers" or "bins" key!')

    axis_binnings = [
        BinningDefinition(
            kde_pdf_axis_name_map_inv[var], num_dict[bin_centers_key][idx])
        for (idx, var) in enumerate(num_dict['vars'])
    ]

    vals = num_dict['pdf_vals']
    if denum_dict is not None:
        # A denumerator is required, so we need to divide the numerator pdf
        # values by the denumerator pdf values, by preserving the correct axis
        # order.
        # Construct the slicing selector for the denumerator pdf values array to
        # match the axis order of the numerator pdf values array.
        selector = []
        for var in num_dict['vars']:
            if var in denum_dict['vars']:
                # The variable is present in both pdf value arrays. So select
                # all values of that dimension.
                selector.append(slice(None, None))
            else:
                # The variable is not present in the normalization pdf value
                # array, so we need to add a dimension for that variable.
                selector.append(np.newaxis)

        denum = denum_dict['pdf_vals'][tuple(selector)]
        denum_nonzero_mask = denum != 0
        out = np.zeros_like(vals)
        np.divide(vals, denum, where=denum_nonzero_mask, out=out)
        vals = out

    # Set infinite values to NaN.
    vals[np.isinf(vals)] = np.nan
    # Set NaN values to 0.
    vals[np.isnan(vals)] = 0

    pdf = multidimgridpdf_cls(
        pmm=pmm,
        axis_binnings=axis_binnings,
        pdf_grid_data=vals,
        norm_factor_func=norm_factor_func,
        cache_pd_values=cache_pd_values,
        **kwargs)

    return pdf



# -*- coding: utf-8 -*-
# Author: Dr. Martin Wolf <mail@martin-wolf.org>

from skyllh.core.dataset import (
    DatasetCollection,
)
from skyllh.i3.dataset import (
    I3Dataset,
)


def create_dataset_collection(
        cfg,
        base_path=None,
        sub_path_fmt=None,
):
    """Defines a dataset collection with a test dataset.

    Parameters
    ----------
    cfg : instance of Config
        The instance of Config holding the local configuration.
    base_path : str | None
        The base path of the data files. The actual path of a data file is
        assumed to be of the structure <base_path>/<sub_path>/<file_name>.
        If None, use the default path ``cfg['repository']['base_path']``.
    sub_path_fmt : str | None
        The sub path format of the data files of the public data sample.
        If None, use the default sub path format
        'testdata'.

    Returns
    -------
    dsc : DatasetCollection
        The dataset collection containing all the seasons as individual
        I3Dataset objects.
    """
    (version, verqualifiers) = (1, dict(p=0))

    default_sub_path_fmt = 'testdata'

    dsc = DatasetCollection('Public Data 10-year point-source')

    dsc.description = r"""
    This dataset collection contains a test dataset which can be used for unit
    tests.
    """

    # Define the common keyword arguments for all data sets.
    ds_kwargs = dict(
        cfg=cfg,
        livetime=None,
        version=version,
        verqualifiers=verqualifiers,
        base_path=base_path,
        default_sub_path_fmt=default_sub_path_fmt,
        sub_path_fmt=sub_path_fmt,
    )

    TestData = I3Dataset(
        name='TestData',
        exp_pathfilenames='exp.npy',
        mc_pathfilenames='mc.npy',
        grl_pathfilenames='grl.npy',
        **ds_kwargs,
    )

    dsc.add_datasets((
        TestData,
    ))

    return dsc



import numpy as np
from scipy.stats import norm


def em_expectation_step(
        ns,
        mu,
        sigma,
        t,
        sob,
):
    """Expectation step of expectation maximization algorithm.

    Parameters
    ----------
    ns : instance of ndarray
        The (n_flares,)-shaped numpy ndarray holding the number of signal
        neutrinos, as weight for each gaussian flare.
    mu : instance of ndarray
        The (n_flares,)-shaped numpy ndarray holding the mean for each gaussian
        flare.
    sigma: instance of ndarray
        The (n_flares,)-shaped numpy ndarray holding the sigma for each gaussian
        flare.
    t : instance of ndarray
        The (n_events,)-shaped numpy ndarray holding the time of each event.
    sob : instance of ndarray
        The (n_events,)-shaped numpy ndarray holding the signal-over-background
        values of each event.

    Returns
    -------
    expectations : instane of ndarray
        The (n_flares, n_events)-shaped numpy ndarray holding the expectation
        of each flare and event.
    llh : float
        The log-likelihood value, which is the sum of log of the signal and
        background expectations.
    """
    n_flares = len(ns)

    b_term = (1 - np.cos(10 / 180 * np.pi)) / 2
    N = len(t)
    e_sig = np.empty((n_flares, N), dtype=np.float64)
    for i in range(n_flares):
        e_sig[i] = norm(loc=mu[i], scale=sigma[i]).pdf(t)
        e_sig[i] *= sob
        e_sig[i] *= ns[i]
    e_bkg = (N - np.sum(ns)) / (np.max(t) - np.min(t)) / b_term
    denom = np.sum(e_sig, axis=0) + e_bkg

    expectations = e_sig / denom
    llh = np.sum(np.log(denom))

    return (expectations, llh)


def em_maximization_step(
        e,
        t,
):
    """The maximization step of the expectation maximization algorithm.

    Parameters
    ----------
    e : instance of ndarray
        The (n_flares, n_events)-shaped numpy ndarray holding the expectation
        for each event and flare.
    t : 1d ndarray of float
        The times of each event.

    Returns
    -------
    mu : list of float
        Best fit mean time of the gaussian flare.
    sigma : list of float
        Best fit sigma of the gaussian flare.
    ns : list of float
        Best fit number of signal neutrinos, as weight for the gaussian flare.
    """
    mu = []
    sigma = []
    ns = []
    for i in range(e.shape[0]):
        mu.append(np.average(t, weights=e[i]))
        sigma.append(np.sqrt(np.average(np.square(t - mu[i]), weights=e[i])))
        ns.append(np.sum(e[i]))
    sigma = [max(1, s) for s in sigma]

    return (mu, sigma, ns)


def em_fit(
        x,
        weights,
        n=1,
        tol=1.e-200,
        iter_max=500,
        weight_thresh=0,
        initial_width=5000,
        remove_x=None,
):
    """Perform the expectation maximization fit.

    Parameters
    ----------
    x : array of float
        The quantity to run EM on (e.g. the time if EM should find time flares).
    weights : array of float
        The weights for each x value (e.g. the signal over background ratio).
    n : int
        How many Gaussians flares we are looking for.
    tol : float
        The stopping criteria for the expectation maximization. This is the
        difference in the normalized likelihood over the last 20 iterations.
    iter_max : int
        The maximum number of iterations, even if stopping criteria tolerance
        (``tol``) is not yet reached.
    weight_thresh : float
        Set a minimum threshold for event weights. Events with smaller weights
        will be removed.
    initial_width : float
        The starting width for the gaussian flare in days.
    remove_x : float | None
        Specific x of event that should be removed.

    Returns
    -------
    mu : list of float
        The list of size ``n`` with the determined mean values.
    sigma : list of float
        The list of size ``n`` with the standard deviation values.
    ns : list of float
        The list of size ``n`` with the normalization factor values.
    """
    if weight_thresh > 0:
        # Remove events below threshold.
        for i in range(len(weights)):
            mask = weights > weight_thresh
            weights[i] = weights[i][mask]
            x[i] = x[i][mask]

    if remove_x is not None:
        # Remove data point.
        mask = x == remove_x
        weights = weights[~mask]
        x = x[~mask]

    # Do the expectation maximization.
    mu = np.linspace(x[0], x[-1], n+2)[1:-1]
    sigma = np.full((n,), initial_width)
    ns = np.full((n,), 10)

    llh_diff = 100
    llh_old = 0
    llh_diff_list = [100] * 20

    # Run until convergence or maximum number of iterations is reached.
    iteration = 0
    while (iteration < iter_max) and (llh_diff > tol):
        iteration += 1

        (e, llh_new) = em_expectation_step(
            ns=ns,
            mu=mu,
            sigma=sigma,
            t=x,
            sob=weights)

        tmp_diff = np.abs(llh_old - llh_new) / llh_new
        llh_diff_list = llh_diff_list[:-1]
        llh_diff_list.insert(0, tmp_diff)
        llh_diff = np.max(llh_diff_list)

        llh_old = llh_new

        (mu, sigma, ns) = em_maximization_step(
            e=e,
            t=x)

    return (mu, sigma, ns)
